
import io
import os
import glob
import json
import collections
import warnings
import logging
import zipfile
import sys
if sys.version_info.minor < 7:
    from importlib_resources import open_text, open_binary
else:
    from importlib.resources import open_text, open_binary

from retrieve.data import Doc, Ref, Collection


logger = logging.getLogger(__name__)


def decode_ref(ref):
    book, chapter, verse = ref.split('_')
    book = ' '.join(book.split('-'))
    return book, chapter, verse


def encode_ref(ref):
    book, chapter, verse = ref
    return '-'.join(book.split()) + '_' + '_'.join([chapter, verse])


def read_testament_books(testament):
    lines = open_text('retrieve.resources.misc', '{}-testament.books'.format(testament)).read()
    books = []
    for line in lines.split('\n'):
        books.append(line.strip())
    return books

read_testament_books.__test__ = False # Tell nose to ignore test


def read_bible(path_or_file, fields=('token', 'pos', '_', 'lemma'),
               max_verses=-1, sort_docs=True, verse_ids=None):

    if isinstance(path_or_file, str):
        f = open(path_or_file)
    else: # assume is a stream
        f = path_or_file

    docs = []
    for line in f:
        line = line.strip()
        if not line:
            continue
        if max_verses > 0 and len(docs) >= max_verses:
            break
        book, chapter, verse, *data = line.split('\t')
        if verse_ids is not None and (book, chapter, verse) not in verse_ids:
            continue
        data = [field.split() for field in data]
        if len(data) != len(fields):
            raise ValueError(
                "Expected {} metadata fields, but got {}."
                .format(len(fields), len(data)))
        doc_id = book, chapter, verse
        try:
            docs.append(Doc(fields=dict(zip(fields, data)), doc_id=doc_id))
        except ValueError as e:
            warnings.warn("Ignoring document {}".format(' '.join(doc_id)))

    if not sort_docs:
        return docs

    books = {book: idx for idx, book in enumerate(
        read_testament_books('old') + read_testament_books('new'))}

    def key(doc):
        book, chapter, verse = doc.doc_id
        return books.get(book, len(books)), int(chapter), int(verse)

    return sorted(docs, key=key)


def read_vulgate(**kwargs):
    zipf = zipfile.ZipFile(open_binary('retrieve.resources.texts', 'vulgate.zip'))
    return read_bible(io.StringIO(zipf.read('vulgate.csv').decode()), **kwargs)


def load_bible(docs, name='bible',
               include_blb=False, split_testaments=False, max_targets=None):

    coll = Collection(docs, name=name)
    if not split_testaments:
        return coll

    # split
    old, new = set(read_testament_books('old')), set(read_testament_books('new'))
    old_books, new_books = [], []
    for doc in docs:
        if doc.doc_id[0] in old:
            old_books.append(doc)
        elif doc.doc_id[0] in new:
            new_books.append(doc)
        else:
            warnings.warn("Missing book: {}".format(doc.doc_id[0]))

    # add refs
    old = Collection(old_books, name='Old Testament')
    new = Collection(new_books, name='New Testament')
    if not include_blb:
        return old, new

    refs = []
    for ref in load_blb_refs(max_targets=max_targets):
        source, target = [], []
        for r in ref['source']:
            if r not in new:
                warnings.warn("Couldn't find verse: new " + str(r))
                continue
            source.append(new.get_doc_idx(r))
        for r in ref['target']:
            if r not in old:
                warnings.warn("Couldn't find verse: old " + str(r))
                continue
            target.append(old.get_doc_idx(r))

        if len(source) == 0 or len(target) == 0:
            warnings.warn("Missing refs for expected pair" + str(ref))
            continue

        refs.append(Ref(tuple(source), tuple(target),
                        meta={'ref_type': ref['ref_type'],
                              'source': ref['source'],
                              'target': ref['target']}))

    return old, new, refs


def load_vulgate(split_testaments=False, max_targets=None, **kwargs):
    return load_bible(
        read_vulgate(**kwargs),
        include_blb=False, split_testaments=split_testaments, max_targets=max_targets)


def load_blb_refs(max_targets=None):
    data = open_text('retrieve.resources.misc', 'blb.refs.json').read()

    output = []
    for ref in json.loads(data):
        source = [decode_ref(s_ref) for s_ref in ref['source']]
        target = [decode_ref(t_ref) for t_ref in ref['target']]
        ref_type = ref['ref_type']
        if max_targets is not None:
            if (len(source) > max_targets or len(target) > max_targets):
                continue
        output.append({'source': source, 'target': target, 'ref_type': ref_type})

    return output


def read_doc(path, fields=('token', 'pos', '_', 'lemma'), sep='\t'):
    output = collections.defaultdict(list)
    with open(path) as f:
        for line in f:
            try:
                data = line.strip().split(sep)
                if len(data) != len(fields):
                    raise ValueError(
                        "Expected {} metadata fields, but got {}. File: {}"
                        .format(len(fields), len(data), path))
                for key, val in zip(fields, data):
                    if key != '_':
                        output[key].append(val)
            except ValueError as e:
                raise e
            except Exception:
                print(line)

    return dict(output)


def read_refs(path):
    with open(path) as f:
        output = []
        for ref in json.load(f):
            ref['target'] = list(map(tuple, ref['target']))
            output.append(ref)
        return output


def shingle_doc(doc, f_id, overlap=10, window=20):
    shingled_docs = []
    n_words = len(doc[next(iter(doc.keys()))])
    for start in range(0, n_words, window - overlap):
        # doc might be smaller than window
        stop = min(start + window, n_words)
        assert start < stop, (start, stop, f_id)
        # doc id
        doc_id = f_id, (start, stop)
        # prepare doc
        fields = {key: vals[start:stop] for key, vals in doc.items()}
        fields['ids'] = list(range(start, stop))

        shingled_docs.append(Doc(fields=fields, doc_id=doc_id))

    return shingled_docs


def load_bernard(directory='data/texts/bernard',
                 max_targets=None,
                 **kwargs):

    bible = Collection(read_vulgate(), name='Vulgate')
    shingled_docs, shingled_refs = [], []
    for path in glob.glob(os.path.join(directory, '*.txt')):
        refs = read_refs(path.replace('.txt', '.refs.json'))
        doc = read_doc(path, fields=('w_id', 'token', 'pos', '_', 'lemma'))
        shingles = shingle_doc(doc, path, **kwargs)
        for ref in refs:
            source = []
            target = ref['target']
            if any(t not in bible for t in target):
                warnings.warn("Missing target")
                continue
            for shingle in shingles:
                if set(ref['source']).intersection(set(shingle.fields['ids'])):
                    source.append(shingle.doc_id)
            if source and (max_targets is None or len(source) <= max_targets) and \
               target and (max_targets is None or len(target) <= max_targets):
                shingled_refs.append(Ref(tuple(source), tuple(target), meta=ref))
            else:
                warnings.warn("Ignoring target")

        shingled_docs.extend(shingles)

    shingled_docs = Collection(shingled_docs, name='Bernard')

    return shingled_docs, bible, shingled_refs




import unittest

import numpy as np
import scipy.sparse

from retrieve.corpora import load_vulgate
from retrieve.data import Criterion, TextPreprocessor, FeatureSelector
from retrieve.embeddings import Embeddings
from retrieve.methods import Tfidf
from retrieve.methods.vsm.soft_cosine import soft_cosine_similarities
from retrieve.methods.vsm.soft_cosine import soft_cosine_simple
from retrieve.methods.vsm.soft_cosine import parallel_soft_cosine

from sklearn.metrics import pairwise_kernels


class TestSoftCosine(unittest.TestCase):
    def setUp(self):
        # load
        vulg = load_vulgate(max_verses=1000)
        # preprocess
        TextPreprocessor().process_collections(vulg, min_n=1, max_n=1)
        # drop features and get vocabulary
        fsel = FeatureSelector(vulg)
        vocab = fsel.filter_collections(
            vulg, criterion=(Criterion.DF >= 2) & (Criterion.FREQ >= 5))
        # get documents
        feats = vulg.get_features(cast=set)
        # transform to tfidf
        feats = Tfidf(vocab).fit(feats).transform(feats)
        query, index = feats[:feats.shape[0]//2], feats[feats.shape[0]//2:]
        # load embeddings, make sure S is in same order as vocab
        embs = Embeddings.from_resource(
            'latin.lemma.ft.dim100.mc2.embeddings.gz', vocab=vocab)

        self.embs = embs
        self.query = query
        self.index = index
        self.vocab = vocab

    def test_degenerate(self):
        # simple cosine similarity (we always return normalized vectors)
        sims1 = pairwise_kernels(self.query, self.index, metric='linear')
        # degenerate soft cosine should be equal to cosine
        sims2 = soft_cosine_similarities(
            self.query, self.index, np.identity(len(self.vocab)))

        self.assertTrue(np.allclose(sims1, sims2))

    def test_full(self):
        S = self.embs.get_S(vocab=self.vocab, fill_missing=True, cutoff=0.75, beta=2)
        sims = soft_cosine_similarities(self.query, self.index, S)
        for i in range(10):
            for j in range(10):
                sim1 = sims[i, j]
                sim2 = soft_cosine_simple(self.query[i], self.index[j], S)
                self.assertAlmostEqual(sim1, sim2, msg=f"{i}!={j}; {sim1}:{sim2}")

    def test_sparse(self):
        S = self.embs.get_S(vocab=self.vocab, fill_missing=True, cutoff=0.75, beta=2)
        self.assertTrue(
            scipy.sparse.issparse(S),
            msg="get_S returns sparse if fill_missing is True or cutoff > 0.0")
        sims2 = soft_cosine_similarities(self.query, self.index, S)
        self.assertTrue(
            scipy.sparse.issparse(sims2),
            msg="soft_cosine_similarities returns sparse if S is sparse")
        sims1 = soft_cosine_similarities(self.query, self.index, S.todense())
        self.assertFalse(
            scipy.sparse.issparse(sims1),
            msg="soft_cosine_similarities returns dense if S is dense")
        self.assertTrue(
            np.allclose(sims2.todense(), sims1),
            msg="sparse and dense results match")

    def test_parallel(self):
        S = self.embs.get_S(vocab=self.vocab, fill_missing=True, cutoff=0.75, beta=2)
        sims_parallel = parallel_soft_cosine(self.query, self.index, S, threshold=0.25)
        sims_single = soft_cosine_similarities(self.query, self.index, S, threshold=0.25)

        self.assertTrue(all(sims_parallel.data == sims_single.data))





"""
Modification from https://github.com/ekzhu/SetSimilaritySearch
LICENSE: https://raw.githubusercontent.com/ekzhu/SetSimilaritySearch/master/LICENSE
"""

import logging
from collections import defaultdict, deque

import numpy as np

from SetSimilaritySearch.utils import _frequency_order_transform, \
        _similarity_funcs, _overlap_threshold_funcs, _position_filter_funcs, \
        _symmetric_similarity_funcs, _asymmetric_similarity_funcs, \
        _overlap_index_threshold_funcs

logger = logging.getLogger(__name__)


class SearchIndex(object):
    """This data structure supports set similarity search queries.
    The algorithm is a combination of the prefix filter and position filter
    techniques.

    Args:
        sets (list): a list of sets, each entry is an iterable representing a
            set.
        similarity_func_name (str): the name of the similarity function used;
            this function currently supports `"jaccard"`, `"cosine"`, and
            `"containment"`.
        similarity_threshold (float): the threshold used, must be a float
            between 0 and 1.0.
    """

    def __init__(self, index, queries=None, similarity_func_name="jaccard",
                 similarity_threshold=0.5):
        if not isinstance(index, list) or len(index) == 0:
            raise ValueError("Input parameter sets must be a non-empty list.")
        if similarity_func_name not in _similarity_funcs:
            raise ValueError("Similarity function {} is not supported.".format(
                similarity_func_name))
        if similarity_threshold < 0 or similarity_threshold > 1.0:
            raise ValueError(
                "Similarity threshold must be in the range [0, 1].")
        self.similarity_threshold = similarity_threshold
        self.similarity_func = _similarity_funcs[similarity_func_name]
        self.overlap_threshold_func = \
            _overlap_threshold_funcs[similarity_func_name]
        self.overlap_index_threshold_func = \
            _overlap_index_threshold_funcs[similarity_func_name]
        self.position_filter_func = _position_filter_funcs[similarity_func_name]
        logger.debug("Building SearchIndex on {} sets.".format(len(index)))
        logger.debug("Start frequency transform.")
        self.sets, self.order = _frequency_order_transform(index + (queries or []))
        self.n_sets = len(index)
        logger.debug("Finish frequency transform, {} tokens in total.".format(
            len(self.order)))
        self.index = defaultdict(list)
        logger.debug("Start indexing sets.")
        for i, s in enumerate(self.sets):
            prefix = self._get_prefix_index(s)
            for j, token in enumerate(prefix):
                self.index[token].append((i, j))
        logger.debug("Finished indexing sets.")

    def _get_prefix_index(self, s):
        t = self.overlap_index_threshold_func(len(s), self.similarity_threshold)
        prefix_size = len(s) - t + 1
        return s[:prefix_size]

    def _get_prefix(self, s):
        t = self.overlap_threshold_func(len(s), self.similarity_threshold)
        # it can be shown that we only need to index a prefix of
        # length|x|−⌈t·|x|⌉+1 for every recordxto ensure the prefix filtering-based
        # method does not miss any similarity join result.
        prefix_size = len(s) - t + 1
        return s[:prefix_size]

    def query(self, s):
        """Query the search index for sets similar to the query set.

        Args:
            s (Iterable): the query set.

        Returns (list): a list of tuples `(index, similarity)` where the index
            is the index of the matching sets in the original list of sets.
        """
        s1 = np.sort([self.order[token] for token in s if token in self.order])
        logger.debug("{} original tokens and {} tokens after applying "
                     "frequency order.".format(len(s), len(s1)))
        prefix = self._get_prefix(s1)
        candidates = set([i for p1, token in enumerate(prefix)
                          for i, p2 in self.index[token]
                          if (i < self.n_sets and
                              self.position_filter_func(s1, self.sets[i], p1, p2,
                                                        self.similarity_threshold))])
        logger.debug("{} candidates found.".format(len(candidates)))
        results = deque([])
        for i in candidates:
            s2 = self.sets[i]
            sim = self.similarity_func(s1, s2)
            if sim < self.similarity_threshold:
                continue
            results.append((i, sim))
        logger.debug("{} verified sets found.".format(len(results)))
        return list(results)




import itertools
import uuid
import functools
import operator
import collections
from typing import Any, List, Dict, Tuple
import string
import re
import logging
from dataclasses import dataclass

import numpy as np

from retrieve import utils
from retrieve.methods import (jaccard, containment, containment_min,
                              weighted_containment, weighted_jaccard, cosine)
from retrieve.methods import local_alignment, get_horizontal_alignment

logger = logging.getLogger(__name__)


@dataclass(eq=True, frozen=True)
class Ref:
    """
    Dataclass to store reference information. It's hashable and allows
    to incorporate metadata in the `meta` field as a tuple
    """
    source: Tuple[Any]
    target: Tuple[Any]
    meta: tuple = ()


def printable_doc_id(doc_id, levels=('-', ':', '+')):
    def _printable_doc_id(doc_id, level):
        if isinstance(doc_id, (list, tuple)):
            for it in doc_id:
                yield from _printable_doc_id(it, level + 1)
        else:
            yield level, str(doc_id)

    last, output, groups = None, None, list(_printable_doc_id(doc_id, -1))
    for level, group in itertools.groupby(groups, key=lambda tup: tup[0]):
        group = levels[level].join(it for _, it in group)
        if output is None:
            output = group
        else:
            output = levels[min(level, last)].join([output, group])
        last = level
    return output


class Doc:
    """
    Convenience class representing a document.

    Keeps any morphological tags passed to the constructor in memory
    as well as doc ids and refs

    Arguments
    =========
    fields : dict containing input
        must include a "token" field with the tokenized text
        currently it also assumes a "lemma" field
        "pos" can be used for further preprocessing
    doc_id : any document identifier
    ref : Ref
    """
    def __init__(self,
                 fields: Dict[str, List[any]],
                 doc_id: Any,
                 ignore_fields=set(['_'])):

        if isinstance(doc_id, int):
            raise ValueError("Can't use `doc_id` of type integer")
        if 'token' not in fields:
            raise ValueError("`fields` requires 'token' data")

        self.doc_id = doc_id
        self.fields = {f: data for f, data in fields.items() if f not in ignore_fields}
        # check lengths
        self._check_fields(self.fields)
        # data
        self._features = None
        self._preprocessed_features = None

    @staticmethod
    def _check_fields(fields):
        length = None
        for field, data in fields.items():
            if length is None:
                length = len(data)
            else:
                if len(data) != length:
                    raise ValueError("Expected {} of {} but got {}".format(
                        length, field, len(data)))

    @property
    def text(self):
        return ' '.join(self.get_features(field='token'))

    def get_printable_doc_id(self):
        return printable_doc_id(self.doc_id)

    def to_counter(self, field=None):
        return collections.Counter(self.get_features(field=field))

    def __repr__(self):
        return '<Doc doc_id={} text="{}"/>'.format(
            str(self.doc_id), self.text[:30] + "...")

    @property
    def feature_density(self):
        if self._preprocessed_features == 0:
            return 0
        return len(self.get_features()) / self._preprocessed_features

    def set_features(self, features):
        self._features = features
        self._preprocessed_features = len(features)

    def get_features(self, field=None):
        if not field:
            if self._features is None:
                raise ValueError("Unprocessed doc: [{}]".format(str(self.doc_id)))
            text = self._features
        else:
            text = self.fields[field]

        return text

    def print_alignment(self, doc, **kwargs):
        a1 = a2 = None
        if 'lemma' in self.fields and 'lemma' in doc.fields:
            a1, a2, _ = local_alignment(self.get_features('lemma'), doc.get_features('lemma'), **kwargs)
        s1, al, s2 = get_horizontal_alignment(
            self.get_features('token'), doc.get_features('token'), a1=a1, a2=a2, **kwargs)
        print(s1 + '\n' + al + '\n' + s2)


def _wrap_fn(fn, use_counter=False):
    def wrapped(this, that, field='lemma', **kwargs):
        if use_counter:
            return fn(this.to_counter(field), that.to_counter(field), **kwargs)
        else:
            return fn(this.get_features(field), that.get_features(field), **kwargs)
    return wrapped


setattr(Doc, 'jaccard', _wrap_fn(jaccard, use_counter=True))
setattr(Doc, 'weighted_jaccard', _wrap_fn(weighted_jaccard, use_counter=True))
setattr(Doc, 'containment', _wrap_fn(containment, use_counter=True))
setattr(Doc, 'containment_min', _wrap_fn(containment_min, use_counter=True))
setattr(Doc, 'weighted_containment', _wrap_fn(weighted_containment, use_counter=True))
setattr(Doc, 'cosine', _wrap_fn(cosine, use_counter=True))
setattr(Doc, 'local_alignment', _wrap_fn(local_alignment))
setattr(Doc, 'get_horizontal_alignment', _wrap_fn(get_horizontal_alignment))


class Collection:
    """
    Class representing a collection of docs

    Arguments
    =========
    docs : list of Doc
    """
    def __init__(self, docs, name=None):
        self._docs = docs
        self._doc_ids = {doc.doc_id: idx for idx, doc in enumerate(docs)}
        # identifier for collection
        self.name = name or str(uuid.uuid4())[:8]
        # processing metadata
        self.preprocesing_summary = None
        self.fsel_summary = None

    def __getitem__(self, idx):
        if isinstance(idx, np.integer):
            idx = int(idx)

        if isinstance(idx, int):
            # access by index
            return self._docs[idx]
        # access by key
        return self._docs[self._doc_ids[idx]]

    def __len__(self):
        return len(self._docs)

    def __contains__(self, doc_id):
        return doc_id in self._doc_ids

    def __iter__(self):
        yield from self.get_docs()

    def get_doc_idx(self, doc_id):
        # access by doc id
        return self._doc_ids[doc_id]

    def get_docs(self, index=None):
        """
        Generator over docs

        Arguments
        =========
        index : list or set or dict of document indices to get (optional)
        """
        # TODO: index should probably be based on doc ids
        if index is not None:
            index = set(index)
        for idx, doc in enumerate(self._docs):
            if index is not None and idx not in index:
                continue
            yield doc

    def get_nonempty_features(self, **kwargs):
        """
        Avoid getting empty bags of features. Documents might
        be empty after preprocessing or feature selection.

        Output
        ======
        Tuple of (features, index):
            features, list of document features
            index, list of indices mapping the document idx in the original collection
        """
        output, index = [], []
        for idx, feats in enumerate(self.get_features(**kwargs)):
            # handle empty documents
            if feats:
                output.append(feats)
                index.append(idx)

        return output, index

    def get_features(self, cast=None, min_features=0, min_feature_density=0, **kwargs):
        """
        Get preprocessed text.

        Arguments
        =========
        cast : func (optonal), document features are casted with `cast` if passed

        Output
        ======
        list of lists with features
        """
        output = []
        for doc in self.get_docs():
            # get features
            feats = doc.get_features(**kwargs)

            # empty input if number of features falls below threshold
            if (min_features > 0 and len(feats) < min_features) or \
               (min_feature_density > 0 and doc.feature_density > min_feature_density):
                feats = []

            if cast is not None:
                feats = cast(feats)

            output.append(feats)
        return output

    def get_field_vocab(self, field):
        """
        Get vocabulary of a given input field (ignores preprocessing)
        """
        return collections.Counter(
            w for doc in self.get_docs() for w in doc.fields[field])

    @classmethod
    def from_file(cls, *paths, 
                  # shingling
                  do_shingling=False, shingle_overlap=10, shingle_window=20):

        docs = []
        for path in paths:
            lines = []
            with open(path) as f:
                for line in f:
                    line = line.strip()
                    if not line:
                        continue
                    lines.append(line.split())
            if do_shingling:
                for doc in shingle_docs(
                        {'token': [tok for l in lines for tok in l]}, path,
                        overlap=shingle_overlap, window=shingle_window):
                    docs.append(doc)
            else:
                d_id = 0
                for line in lines:
                    docs.append(Doc({'token': line}, doc_id=(path, d_id)))
                    d_id += 1

        return cls(docs)

    @classmethod
    def from_csv(cls, *paths, drop_diacritics=None, 
                 fields=('token', 'lemma', 'pos'), sep='\t', read_header=False,
                 ignore_empty_lines=False, 
                 # shingling
                 do_shingling=False, shingle_overlap=10, shingle_window=20):
        """
        Arguments
        =========

        Create collection from space-separated csv like file

        Sic sic PROPN
        itur itur PROPN
        ad ad PROPN
        astra astra PROPN
        . . PUNCT

        En en ADP
        una uno DET
        mañana mañana NOUN
        de de ADP
        Diciembre diciembre NOUN

        using sentence numbers as doc ids
        """
        docs = []
        for path in paths:
            d_id = 0
            for doc in read_csv(path,
                    fields=fields, sep=sep, read_header=read_header, 
                    ignore_empty_lines=ignore_empty_lines):
                if do_shingling:
                    for shingle in shingle_docs(
                            doc, path, overlap=shingle_overlap, window=shingle_window):
                        docs.append(shingle)
                else:
                    docs.append(Doc(doc, doc_id=(path, d_id)))
                    d_id += 1

        return cls(docs)


class TextPreprocessor:
    """
    Preprocess docs based on doc metadata
    """

    PUNCT = r"[{}]+".format(string.punctuation)

    def __init__(self,
                 field='lemma',
                 lower=True,
                 field_regexes={},
                 drop_punctuation=True, punct_field='token',
                 replace_unk=False, drop_unk=False, unk_token='$unk$', unk_field='token',
                 stopwords=None, stop_field='lemma'):
        """
        field : str, what field to use to extract features
        lower : bool, whether to lowercase input text
        field_regexes : dict (field : regex), a dict mapping fields to regexes that tokens
            should match in order to not be filtered out
        drop_punctuation : bool, whether to drop punctuation
        punct_field : str, field that should be used to check for punctuation
        replace_unk : bool, whether to replace unknown lemmas with the corresponding token
            from the field specified by `unk_field`
        drop_unk : bool, whether to remove tokens with unknown lemmas
        unk_token : str, token used to represent unknown lemmas in the lemma field
        unk_field : str, field to use for the replacement of the unknown lemma 
            if `replace_unk` is True.
        stopwords : retrieve.utils.Stopwords, stopwords to filter out
        stop_field : str, field to be used for checking stopwords
        """
        self.field = field
        self.lower = lower
        # punctuation
        self.drop_punctuation = drop_punctuation
        self.punct_field = punct_field
        # unks
        self.replace_unk = replace_unk
        self.drop_unk = drop_unk
        self.unk_token = unk_token
        self.unk_field = unk_field
        # stopwords
        self.stopwords = stopwords
        self.stop_field = stop_field
        # regexes
        self.field_regexes = field_regexes

    def get_summary(self):
        return dict({
            'field': self.field,
            'lower': self.lower,
            # punctuation
            'drop_punctuation': self.drop_punctuation,
            'punct_field': self.punct_field,
            # unks
            'replace_unk': self.replace_unk,
            'drop_unk': self.drop_unk,
            'unk_token': self.unk_token,
            'unk_field': self.unk_field,
            # stopwords
            'stopwords': self.stopwords.get_summary() if self.stopwords else None,
            'stop_field': self.stop_field,
            # regexes
            'field_regexes': self.field_regexes
        })

    def process_generator(self, doc):
        for i in range(len(doc.fields['token'])):
            target = doc.fields[self.field][i]

            if self.drop_punctuation and re.fullmatch(
                    TextPreprocessor.PUNCT, doc.fields[self.punct_field][i]):
                logger.debug("Dropping punctuation: {}".format(
                    doc.fields[self.punct_field][i]))
                continue

            if self.stopwords is not None:
                if doc.fields[self.stop_field][i].lower() in self.stopwords:
                    logger.debug("Dropping stopword: {}".format(
                        doc.fields[self.stop_field][i]))
                    continue

            reg_match = True
            for re_field, regex in self.field_regexes.items():
                if not re.match(regex, doc.fields[re_field][i]):
                    reg_match = False
                    break
            if not reg_match:
                logger.debug("Dropping regex {}: {}".format(
                    re_field, doc.fields[re_field][i]))
                continue

            if (self.replace_unk or self.drop_unk):
                if 'lemma' in doc.fields and doc.fields['lemma'][i] == self.unk_token:
                    if self.replace_unk:
                        target = doc.fields[self.unk_field][i]
                    elif self.drop_unk:
                        logger.debug("Dropping unknown")
                        continue

            if self.lower:
                target = target.lower()

            yield target

    def process(self, doc, ngram_extractor=None):
        """
        Process input text creating n-grams on the processed output.
        """
        processed = self.process_generator(doc)
        if ngram_extractor is not None:
            yield from ngram_extractor.get_ngrams_generator(list(processed))
        else:
            yield from processed

    def process_collections(self, *colls, post_fn=None, **kwargs):
        """
        Process entire collections

        post_fn : function that takes generator of ngrams and produces a new generator
        kwargs : Ngrams additional arguments
        """
        ngram_extractor = utils.Ngrams(**kwargs)
        summary = self.get_summary()
        summary['ngrams'] = ngram_extractor.get_summary()
        for coll in colls:
            coll.preprocesing_summary = summary
            for doc in coll.get_docs():
                features = self.process(doc, ngram_extractor=ngram_extractor)
                if post_fn is not None:
                    features = post_fn(features)
                doc.set_features(list(features))


class MetaCriterion(type):
    @property
    def DF(cls):
        return cls("DF")

    @property
    def FREQ(cls):
        return cls("FREQ")

    @property
    def IDF(cls):
        return cls("IDF")


class Criterion(object, metaclass=MetaCriterion):
    def __init__(self, field):
        self.field = field
        self.ops = []
        self.fields_, self.ops_ = [], []

    def _get_index(self, stats, val, operator):
        if val < 1 or (val == 1 and isinstance(val, float)):
            # convert stats to normalized ranks
            # assumes stats has already been argsorted
            stats = np.linspace(0, 1, len(stats))[stats.argsort().argsort()]

        index, = np.where(operator(stats, val))
        return index

    def get_fields_and_ops(self):
        # concat current field to extra fields
        fields = [self.field] + self.fields_
        ops = [self.ops] + self.ops_
        return fields, ops

    def apply(self, f_sel):
        if not self.ops:
            raise ValueError("Criterion not set until comparison")

        fields, ops = self.get_fields_and_ops()
        stats = {f: f_sel._get_stats(f) for f in set(fields)}

        index = []
        for field, ops in zip(fields, ops):
            for val, op in ops:
                index.append(self._get_index(stats[field], val, op))

        if len(index) == 1:
            index = index[0]
        else:
            index = functools.reduce(np.intersect1d, index)

        return index

    def __and__(self, other):
        if not self.ops or not other.ops:
            raise ValueError("Criterion not set until comparison")

        self.fields_ += [other.field]
        self.ops_ += [other.ops]
        return self

    def __le__(self, val):
        self.ops.append((val, operator.le))
        return self

    def __lt__(self, val):
        self.ops.append((val, operator.lt))
        return self

    def __ge__(self, val):
        self.ops.append((val, operator.ge))
        return self

    def __gt__(self, val):
        self.ops.append((val, operator.gt))
        return self

    def __eq__(self, val):
        self.ops.append((val, operator.eq))
        return self


class FeatureSelector:
    def __init__(self, *colls):
        self.fitted = False
        self.features = {}
        self.freqs = []
        self.dfs = []
        self.ndocs = 0
        self.register(*colls)

    def register_text(self, text):
        for ft, cnt in collections.Counter(text).items():
            idx = self.features.get(ft, len(self.features))
            if ft not in self.features:
                self.features[ft] = idx
                self.freqs.append(cnt)
                self.dfs.append(1)
            else:
                self.freqs[idx] += cnt
                self.dfs[idx] += 1
        self.ndocs += 1

    def register(self, *colls):
        for coll in colls:
            for feats in coll.get_features():
                self.register_text(feats)
        self._fit()

        return self

    def _fit(self):
        assert len(self.features) == len(self.freqs) == len(self.dfs)
        self.freqs, self.dfs = np.array(self.freqs), np.array(self.dfs)
        self.fitted = True

    def _get_stats(self, field):
        if not self.fitted:
            raise ValueError("Selector isn't fitted")

        if field == "FREQ":
            return self.freqs
        elif field == "DF":
            return self.dfs
        elif field == "IDF":
            return np.log(np.sum(self.freqs) / self.dfs)
        else:
            raise ValueError("Requested unknown stats")

    def get_vocab(self, criterion=None, return_summary=False):
        if not self.fitted:
            raise ValueError("Selector isn't fitted")

        id2ft = {idx: ft for ft, idx in self.features.items()}

        if criterion is None:
            vocab = {id2ft[idx]: self.freqs[idx] for idx in id2ft}

        else:
            index = criterion.apply(self)
            index = sorted(index, key=self.freqs.__getitem__, reverse=True)
            vocab = {id2ft[idx]: self.freqs[idx] for idx in index}

        if return_summary:
            return vocab, criterion.get_fields_and_ops() if criterion else None

        return vocab

    def filter_collections(self, *colls, criterion=None):
        vocab, summary = self.get_vocab(criterion, return_summary=True)
        for coll in colls:
            coll.fsel_summary = summary
            for doc in coll.get_docs():
                doc.set_features(list(filter(vocab.get, doc.get_features())))

        return vocab

    def filter_texts(self, texts, criterion):
        vocab = self.get_vocab(criterion)
        for text in texts:
            yield [ft for ft in text if ft in vocab]


def read_csv(path, drop_diacritics=None,
             fields=('token', 'pos', '_', 'lemma'), sep='\t', 
             read_header=False, ignore_empty_lines=False):
    """
    Read sentences from csv file. Assumes that empty lines correspond to sentence
    boundaries. This functionality can be turned off with `ignore_empty_lines`.
    The function `shingle_docs` can be used to segment a dataset read
    with this function if no sentence segmentation is present in the csv file.

    Arguments
    =========

    path : str, path to file
    drop_diacritics : iterable or None (optional), fields of the csv file for which
        diacritics should be ignored.
    fields : tuple (optional), field names corresponding to the columns in the csv.
        Ignored if `read_header` is True.
    sep : str, separator
    read_header : bool, whether to use the header to read the fields
    """
    output = []
    c_id = 0

    with open(path) as f:
        if read_header:
            fields = next(f).strip().split(sep)
        if 'token' not in fields:
            raise ValueError("At least field `token` must be available")
        if drop_diacritics is not None:
            if isinstance(drop_diacritics, str):
                drop_diacritics = [drop_diacritics]
            drop_diacritics = set(drop_diacritics)

        sent = collections.defaultdict(list)
        for line in f:
            line = line.strip()
            if not line:
                if ignore_empty_lines:
                    continue
                elif sent:
                    output.append(dict(sent))
                    sent = collections.defaultdict(list)
                    continue
                else:
                    # unexpected empty line at the beginning
                    continue

            data = line.split(sep)
            if len(data) != len(fields):
                raise ValueError(
                    "Expected {} metadata fields, but got {}. File: {}"
                    .format(len(fields), len(data), path))

            for key, val in zip(fields, data):
                if key == '_':
                    continue
                if drop_diacritics is not None and key in drop_diacritics:
                    val = utils.drop_string_diacritics(val)
                sent[key].append(val)

    return output


def shingle_docs(doc, f_id, overlap=10, window=20):
    """
    Produce shingles from a given input dictionary with fields.

    Arguments
    =========
    doc : Doc, instance of a (possibly very long) document
    f_id : str, identifier for the input file, for example, the input file path
    overlap : int, the overlap in tokens over consecutive documents
    window : int, the size of the shingled documents
    """

    output = []
    n_words = len(doc[next(iter(doc.keys()))])
    for start in range(0, n_words, window - overlap):
        # doc might be smaller than window
        stop = min(start + window, n_words)
        assert start < stop, (start, stop, f_id)
        # doc id
        doc_id = f_id, (start, stop)
        # prepare doc
        fields = {key: vals[start:stop] for key, vals in doc.items()}
        fields['ids'] = list(range(start, stop))

        output.append(Doc(fields=fields, doc_id=doc_id))

    return output



"""
Inspired by:
 - https://github.com/biocore/scikit-bio/blob/master/skbio/alignment/_pairwise.py
 - https://github.com/lingpy/lingpy/blob/master/lingpy/algorithm/cython/calign.pyx
"""

import logging

import itertools
from itertools import product

import numba as nb
import numpy as np

try:
    import _align
except ModuleNotFoundError:
    import warnings
    warnings.warn("Couldn't find cython 'align' extension, defaulting to numba implementation")
    _align = None


logger = logging.getLogger(__name__)

VGAP, HGAP, END, MATCH, UNPROCESSED = range(1, 6)
GAP_SYM = -1


def _get_embedding_scores(s1, s2, d, S, match, mismatch):
    scores = np.zeros((len(s1), len(s2)))
    for (i, a), (j, b) in product(enumerate(s1), enumerate(s2)):
        if a == b:
            scores[i, j] = match
        elif a in d and b in d:
            scores[i, j] = match * S[d[a], d[b]]
        else:
            scores[i, j] = mismatch
    return scores


def create_embedding_scorer(embs, match=2, mismatch=-1, **kwargs):
    """
    embs : retrieve.Embeddings object
    """
    if 'fill_missing' in kwargs:
        raise ValueError("`fill_missing` not allowed for EmbeddingScorer")
    keys, S = embs.get_S(**kwargs)
    return EmbeddingScorer(keys, S, match=match, mismatch=mismatch)


class EmbeddingScorer:
    def __init__(self, d, S, match=2, mismatch=-1):
        self.match = match
        self.mismatch = mismatch
        self.S = S
        self.d = d
        self.max_s = S.max()

    def get_scores(self, s1, s2):
        return _get_embedding_scores(s1, s2, self.d, self.S, self.match, self.mismatch)

    def get_maximum_score(self, s1, s2):
        return self.match * self.max_s * min(len(s1), len(s2))


def _get_constant_scores(s1, s2, match, mismatch):
    scores = np.zeros((len(s1), len(s2)))
    for (i, a), (j, b) in product(enumerate(s1), enumerate(s2)):
        scores[i, j] = match[i] if a == b else mismatch
    return scores


class ConstantScorer:
    def __init__(self, match=2, mismatch=-1, manual_scores=None):
        self.match = match
        self.mismatch = mismatch
        self.manual_scores = manual_scores

    def get_scores(self, s1, s2):
        # matches can depend on the actual word (e.g. stop-words)
        if self.manual_scores is not None:
            match = [self.manual_scores.get(w, self.match) for w in s1]
        else:
            match = len(s1) * [self.match]
        return _get_constant_scores(s1, s2, match, self.mismatch)

    def get_maximum_score(self, s1, s2):
        return self.match * min(len(s1), len(s2))


@nb.njit
def compute_input_matrices(
        s1len, s2len, scores,
        extend_gap=-1, open_gap=-1, terminal_gap=0):
    smatrix = np.zeros((s1len + 1, s2len + 1))
    tmatrix = np.zeros((s1len + 1, s2len + 1), dtype=np.int64) + END
    best_score = 0.0
    row = col = None

    for i in range(1, s1len + 1):
        for j in range(1, s2len + 1):
            mscore = smatrix[i-1, j-1] + scores[i-1, j-1]

            # vertical gap
            if i == s1len:
                # don't penalize vertical gaps if we are finished with s1
                vscore = smatrix[i-1, j] + terminal_gap
            elif tmatrix[i-1, j] == VGAP:
                vscore = smatrix[i-1, j] + extend_gap
            else:
                vscore = smatrix[i-1, j] + open_gap

            # horizontal gap
            if j == s2len:
                # don't penalize horizontal gaps if we are finished with s2
                hscore = smatrix[i, j-1] + terminal_gap
            elif tmatrix[i, j-1] == HGAP:
                hscore = smatrix[i, j-1] + extend_gap
            else:
                hscore = smatrix[i, j-1] + open_gap

            # pick best move
            score, move = 0.0, 0
            for cscore, cmove in [(0.0, END),
                                  (hscore, HGAP),
                                  (vscore, VGAP),
                                  (mscore, MATCH)]:
                if cscore >= score:
                    score, move = cscore, cmove

            smatrix[i, j], tmatrix[i, j] = score, move

            if score >= best_score:
                best_score = score
                row, col = i, j

    return smatrix, tmatrix, row, col, best_score


@nb.njit
def traceback(smatrix, tmatrix, row, col, gap_sym=GAP_SYM):
    a1, a2 = [], []
    move = tmatrix[row, col]

    while move != END:
        if move == MATCH:
            a1.append(row - 1)
            a2.append(col - 1)
            row -= 1
            col -= 1
        elif move == VGAP:
            a1.append(row - 1)
            a2.append(gap_sym)
            row -= 1
        else:
            a1.append(gap_sym)
            a2.append(col - 1)
            col -= 1
        move = tmatrix[row, col]

    return a1[::-1], a2[::-1]


def get_alignment_string(s1, s2, a1, a2, gap_sym=GAP_SYM):
    return ([s1[idx] if idx != gap_sym else gap_sym for idx in a1],
            [s2[idx] if idx != gap_sym else gap_sym for idx in a2])


def get_alignment_ranges(a1, a2, gap_sym=GAP_SYM):

    def _get_alignment_ranges(alignment, gap_sym):
        ranges = []
        for key, group in itertools.groupby(alignment, lambda item: item == gap_sym):
            if key:
                continue
            group = list(group)
            if len(group) > 1:
                start, *_, end = group
            else:
                start = end = group[0]
            ranges.append((start, end + 1))
        return ranges

    return _get_alignment_ranges(a1, gap_sym), _get_alignment_ranges(a2, gap_sym)


def get_printable_string(s1, s2, a1, a2, gap_sym=GAP_SYM):
    a1s, a2s = get_alignment_string(s1, s2, a1, a2, gap_sym=gap_sym)
    assert len(a1s) == len(a2s)

    max1, max2 = max(map(len, a1s)), max(map(len, a2s))
    lines = []
    for a1_i, a2_i in zip(a1s, a2s):
        if a1_i == a2_i:
            sep = '=='
        elif a1_i == gap_sym or a2_i == gap_sym:
            sep = '-'
        else:
            sep = '!='
        lines.append("{} {} {}".format(
            a1_i.ljust(max1),
            sep,
            a2_i.ljust(max2)))

    return '\n'.join(lines)


def get_horizontal_alignment(s1, s2, a1=None, a2=None, gap_sym=GAP_SYM, **kwargs):
    if a1 is None or a2 is None:
        a1, a2, _ = local_alignment(s1, s2, **kwargs)

    # empty alignment
    if not a1 or not a2 or len(s1) < 2:
        return ' '.join(s1), ' ', ' '.join(s2)

    # find start of both seqs
    str1 = str2 = alignment = ''
    start1 = next(filter(lambda it: it != gap_sym, a1))
    start2 = next(filter(lambda it: it != gap_sym, a2))
    str1 += ' '.join(s1[:start1])
    str2 += ' '.join(s2[:start2])
    str1, str2 = str1.rjust(len(str2)), str2.rjust(len(str1))
    alignment += ' ' * len(str1)
    for w1, w2 in zip(a1, a2):
        if w1 == gap_sym:
            str1 += ' ' + (' ' * len(s2[w2]))
            str2 += ' ' + s2[w2]
            alignment += ' ' + (' ' * len(s2[w2]))
        elif w2 == gap_sym:
            str1 += ' ' + s1[w1]
            str2 += ' ' + (' ' * len(s1[w1]))
            alignment += ' ' + (' ' * len(s1[w1]))
        else:
            str1 += ' ' + s1[w1].rjust(len(s2[w2]))
            str2 += ' ' + s2[w2].rjust(len(s1[w1]))
            sym = ('-' if s1[w1] == s2[w2] else 'x')
            alignment += ' ' + (sym * len(s2[w2])).rjust(len(s1[w1]))

    # attach end of both seqs
    end1 = next(filter(lambda it: it != gap_sym, reversed(a1))) + 1
    end2 = next(filter(lambda it: it != gap_sym, reversed(a2))) + 1
    end1, end2 = ' '.join(s1[end1:]), ' '.join(s2[end2:])
    str1 += ' ' + end1.ljust(len(end2))
    str2 += ' ' + end2.ljust(len(end1))

    return str1, alignment, str2


def get_local_alignment_numba(
        s1, s2, scorer, extend_gap, open_gap, terminal_gap, only_score):
    scores = scorer.get_scores(list(s1), list(s2))
    smatrix, tmatrix, row, col, score = compute_input_matrices(
        len(s1), len(s2), scores, extend_gap, open_gap, terminal_gap)
    if only_score:
        return score
    a1, a2 = traceback(smatrix, tmatrix, row, col)
    return a1, a2, score


def get_local_alignment_cython(
        s1, s2, scorer, extend_gap, open_gap, terminal_gap, only_score):
    scores = scorer.get_scores(list(s1), list(s2))
    return _align.sw_alignment(
        len(s1), len(s2), extend_gap, open_gap, terminal_gap, scores, only_score)


def local_alignment(s1, s2, scorer=ConstantScorer(),
                    extend_gap=-1, open_gap=-1, terminal_gap=0,
                    impl='cython', only_score=False, normalize=False):
    if len(s1) == 0 or len(s2) == 0:
        if only_score:
            return 0.0
        return [], [], 0.0

    if impl == 'cython' and _align is not None:
        fn = get_local_alignment_cython
    else:
        fn = get_local_alignment_numba

    res = fn(s1, s2, scorer, extend_gap, open_gap, terminal_gap, only_score)

    if normalize:
        max_score = scorer.get_maximum_score(s1, s2)
        if only_score:
            return res / max_score
        else:
            a, b, score = res
            return a, b, score / max_score

    return res





import unittest
import multiprocessing as mp

from retrieve import utils
from retrieve.methods import local_alignment, align_collections
from retrieve.corpora import load_vulgate


class TestAlign(unittest.TestCase):
    def test_numba(self):
        s1 = 'AGCACACA'
        s2 = 'ACACACTA'
        self.assertEqual(local_alignment(s1, s2), local_alignment(s1, s2, impl='numba'))

        s1 = "ATAGACGACATACAGACAGCATACAGACAGCATACAGA"
        s2 = "TTTAGCATGCGCATATCAGCAATACAGACAGATACG"
        self.assertEqual(local_alignment(s1, s2), local_alignment(s1, s2, impl='numba'))

    def test_parallel(self):
        coll = load_vulgate(max_verses=500)
        sims_single = align_collections(coll, processes=1, field='lemma')
        cpus = mp.cpu_count()
        with utils.timer() as timer:
            sims_parallel = align_collections(coll, processes=cpus, field='lemma')
            timer(desc='Parallel alignment on {} CPUs'.format(cpus))
        self.assertTrue(all(sims_single.data == sims_parallel.data))




from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import TruncatedSVD

from .base import VSM, VSMSoftCosine, init_sklearn_vectorizer


class Tfidf(VSMSoftCosine):
    def __init__(self, vocab, **kwargs):
        super().__init__(vocab, TfidfVectorizer, **kwargs)

    def fit(self, sents):
        self.vectorizer.fit(' '.join(s) for s in sents)
        return self

    def transform(self, sents):
        return self.vectorizer.transform(' '.join(s) for s in sents)


class BOW(VSMSoftCosine):
    def __init__(self, vocab, **kwargs):
        super().__init__(vocab, CountVectorizer, **kwargs)

    def fit(self, sents):
        self.vectorizer.fit(' '.join(s) for s in sents)
        return self

    def transform(self, sents):
        return self.vectorizer.transform(' '.join(s) for s in sents)


class LSI(VSM):
    def __init__(self, npc, vocab, vectorizer=CountVectorizer, **kwargs):
        self.npc = npc
        self.vectorizer = init_sklearn_vectorizer(
            vectorizer, vocabulary=vocab, **kwargs)
        self.svd = TruncatedSVD(n_components=npc)

    def fit(self, sents):
        self.vectorizer.fit(' '.join(s) for s in sents)
        return self

    def transform(self, sents):
        return self.svd.fit_transform(
            self.vectorizer.transform(
                ' '.join(s) for s in sents))




from .align import EmbeddingScorer, ConstantScorer, create_embedding_scorer
from .align import get_horizontal_alignment, local_alignment
from .align import get_alignment_ranges, get_alignment_string
from .parallel_align import align_collections




import logging
import collections

import numpy as np
from numpy.lib.function_base import _sinc_dispatcher
import scipy.sparse as sparse

from retrieve import utils, sparse_utils
from retrieve.data import Criterion, TextPreprocessor, FeatureSelector
from retrieve.embeddings import Embeddings
from retrieve.methods import SetSimilarity, Tfidf, align_collections
from retrieve.methods import create_embedding_scorer, ConstantScorer


logger = logging.getLogger(__name__)


def require_embeddings(embs, msg='', **kwargs):
    if isinstance(embs, str):
        embs = Embeddings.from_file(embs, **kwargs)
    if not isinstance(embs, Embeddings):
        raise ValueError(msg)
    return embs


def get_vocab_from_colls(*colls, field=None):
    output = collections.Counter()
    for coll in colls:
        for doc in coll:
            output.update(doc.get_features(field=field))
    output, _ = zip(*output.most_common())
    return output


class Match:
    def __init__(self, doc1, doc2, sim):
        self.doc1 = doc1
        self.doc2 = doc2
        self.sim = sim

    def __repr__(self):
        return 'Similarity -> {:.5f}\n\t{}: {}\n\t{}: {}'.format(
            self.sim,
            self.doc1.doc_id, self.doc1.text,
            self.doc2.doc_id, self.doc2.text)


class Results:
    def __init__(self, sims, coll1, coll2):
        self.sims = sparse.csr_matrix(sims)
        self.coll1 = coll1
        self.coll2 = coll2

    @property
    def nnz(self):
        return self.sims.nnz

    def keep_top_k(self, k):
        if self.nnz < k:
            logger.info("Nothing to drop from similarity matrix")
            return

        k_val = np.argsort(self.sims)[-k]
        sparse_utils.set_threshold(self.sims, k_val)

    def drop_sims(self, min_sim):
        sparse_utils.set_threshold(self.sims, min_sim)

    def get_top_matches(self, n=None, min_sim=0, max_sim=None, sample=False):
        x, y, _ = sparse.find(self.sims > min_sim)
        score = self.sims[x, y]
        # find returns a scipy matrix instead of a np array
        score = np.array(score)[0]
        if max_sim is not None:
            index, = np.where(score <= max_sim)
            x, y, score = x[index], y[index], score[index]
        index = np.argsort(score)[::-1]
        if sample and n is not None:
            index = np.random.choice(np.arange(len(index)), size=n, replace=False)
        n = n or len(score)
        for i in index[:n]:
            yield Match(self.coll1[x[i]], self.coll2[y[i]], score[i])


def pipeline(coll1, coll2=None,
             # Text Preprocessing
             field='lemma', lower=True, stopwords=None, stop_field='lemma',
             # Ngrams
             min_n=1, max_n=1, skip_k=0, sep='--',
             # Feature Selection
             criterion=None,
             method='set-based', threshold=0, processes=-1, embs=None, chunk_size=5000,
             # Set-based
             # - SetSimilarity: similarity_fn
             #     ('containment', 'containment_min', 'jaccard')
             # VSM-based
             # - Tfidf: vocab, **sklearn,feature_extraction.text.TfidfVectorizer
             # Alignment-based
             # - match, mismatch, open_gap, extend_gap, cutoff, beta
             method_params={},
             # Soft_cosine_params: cutoff, beta
             use_soft_cosine=False, soft_cosine_params={},
             # whether to use parallel soft-cosine (could run into memory issues)
             parallel_soft_cosine=False,
             # For Blast-style alignment
             precomputed_sims=None,
             # return time stats
             return_stats=False, verbose=False):

    colls = [coll for coll in [coll1, coll2] if coll is not None]

    if isinstance(stopwords, str):
        stopwords = utils.Stopwords(stopwords)

    stats = {}

    with utils.timer() as timer:
        # preprocessing
        TextPreprocessor(
            field=field, lower=lower, stopwords=stopwords, stop_field=stop_field,
        ).process_collections(
            *colls, min_n=min_n, max_n=max_n, skip_k=skip_k, sep=sep)
        fsel = FeatureSelector(*colls)
        # get selected vocabulary
        vocab = fsel.filter_collections(*colls, criterion=criterion)

        stats['preprocessing'] = timer(desc='Preprocessing')

        # similarities
        # - set-based method
        if method.startswith('set'):
            coll1_feats = coll1.get_features(cast=set)
            coll2_feats = coll2.get_features(cast=set) if coll2 else coll1_feats
            sims = SetSimilarity(threshold, **method_params).get_similarities(
                coll1_feats, coll2_feats, processes=processes)

        # - vsm-based method
        elif method.startswith('vsm'):
            coll1_feats = coll1.get_features()
            coll2_feats = coll2.get_features() if coll2 is not None else coll1_feats
            tfidf = Tfidf(vocab, **method_params).fit(coll1_feats + coll2_feats)
            if use_soft_cosine:
                embs = require_embeddings(
                    embs, vocab=get_vocab_from_colls(coll1, coll2, field=field),
                    msg='soft cosine requires embeddings')
                sims = tfidf.get_soft_cosine_similarities(
                    coll1_feats, coll2_feats, embs=embs,
                    threshold=threshold,
                    chunk_size=chunk_size, parallel=parallel_soft_cosine,
                    **soft_cosine_params)
            else:
                sims = tfidf.get_similarities(
                    coll1_feats, coll2_feats, threshold=threshold)

        # - alignment-based
        elif method.startswith('alignment'):
            # get scorer
            if 'scorer' in method_params:
                scorer = method_params['scorer']
            elif embs is not None:
                vocab = get_vocab_from_colls(coll1, coll2, field=field)
                scorer = create_embedding_scorer(
                    require_embeddings(embs, vocab=vocab),
                    vocab=vocab,  # in case embeddings are already loaded
                    **{key: val for key, val in method_params.items()
                       if key in set(['match', 'mismatch', 'cutoff', 'beta'])})
            else:
                scorer = ConstantScorer(
                    **{key: val for key, val in method_params.items()
                       if key in set(['match', 'mismatch'])})

            if precomputed_sims is not None:
                logger.info("Computing {} alignments...".format(precomputed_sims.nnz))
            sims = align_collections(
                coll1, coll2,
                S=precomputed_sims, field=field, processes=processes, scorer=scorer,
                **{key: val for key, val in method_params.items()
                   if key in set(['extend_gap', 'open_gap'])})
        else:
            raise ValueError("Unknown method", method)

        stats['similarity'] = timer(desc='Similarity')

    if return_stats:
        return sims, stats

    return sims




import unittest

import numpy as np
import scipy.sparse

from retrieve.corpora import load_vulgate
from retrieve.data import Criterion, TextPreprocessor, FeatureSelector
from retrieve import sparse_utils
from retrieve.methods.set_similarity import SetSimilarity


class TestSparse(unittest.TestCase):
    def setUp(self):
        old, new = load_vulgate(split_testaments=True)
        TextPreprocessor().process_collections(old, new, min_n=2, max_n=4)
        FeatureSelector(old, new).filter_collections(
            old, new, criterion=(Criterion.DF >= 2) & (Criterion.FREQ >= 5))
        # speed up testing
        self.old = old.get_features(cast=set)[:1000]
        self.new = new.get_features(cast=set)[:1000]

    def test_threshold(self):
        for fn in ["containment", "jaccard", "containment_min"]:
            sims = SetSimilarity(0.2, fn).get_similarities(self.new, self.old)
            for th in np.linspace(0.2, 1, 100):
                x1, y1, _ = scipy.sparse.find(sims >= th)
                sims = sparse_utils.set_threshold(sims, th)
                x2, y2, _ = scipy.sparse.find(sims)
                self.assertEqual(x1.shape, x2.shape)
                self.assertEqual(y1.shape, y2.shape)
                self.assertTrue(np.alltrue(x1 == x2))
                self.assertTrue(np.alltrue(y1 == y2))




import collections
import numpy as np

try:
    import pyemd
except ModuleNotFoundError:
    # pyemd is not installed
    pass


def get_wmd_histograms(s1, s2, w2i):
    w1, weights1 = zip(*[(w2i[w], c) for w, c in collections.Counter(s1).most_common()
                         if w in w2i])
    w2, weights2 = zip(*[(w2i[w], c) for w, c in collections.Counter(s2).most_common()
                         if w in w2i])
    words = sorted(list(set(w1 + w2)))
    indexer = {w: idx for idx, w in enumerate(words)}
    h1, h2 = np.zeros(len(indexer)), np.zeros(len(indexer))
    for w, weight in zip(w1, weights1):
        h1[indexer[w]] = weight / sum(weights1)
    for w, weight in zip(w2, weights2):
        h2[indexer[w]] = weight / sum(weights2)
    return h1, h2, np.array(words)


def get_wmd(s1, s2, dists, w2i, get_flow=False):
    """
    Get WMD for two input sentences
    """
    s1, s2 = s1.split(), s2.split()
    h1, h2, words = get_wmd_histograms(s1, s2, w2i)
    D = dists[np.ix_(words, words)]

    if get_flow:
        return pyemd.emd_with_flow(h1, h2, D)

    return pyemd.emd(h1, h2, D)




# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())


def add_handler_level(level=logging.DEBUG):
    """
    Inspired by urllib3 handling of logging
    """
    fmt = '%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s'
    datefmt = '%Y-%m-%d:%H:%M:%S'
    name = "retrieve-handler"

    logger = logging.getLogger(__name__)
    for handler in logger.handlers:
        if handler.get_name() == name:
            logger.removeHandler(handler)

    handler = logging.StreamHandler()
    handler.set_name(name)
    handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
    logger.addHandler(handler)
    logger.setLevel(level)


def enable_log_level(level=logging.DEBUG):
    add_handler_level(level=level)


from . import utils
from .pipeline import pipeline, Results
from .embeddings import Embeddings





def sw_alignment(
        int s1len,
        int s2len,
        # penalties
        float e_gap,
        float o_gap,
        float t_gap,
        object scores,
	bint only_score):

    # pointers
    cdef int i, j, row, col, move
    cdef float mscore, vscore, hscore, best_score
    cdef int VGAP = 1
    cdef int MATCH = 2
    cdef int HGAP = 3
    cdef int END = 4
    cdef int GAP_SYM = -1

    # output and intermediate computations
    cdef list a1 = []
    cdef list a2 = []
    cdef list smatrix = [[0.0 for i in range(s2len + 1)] for j in range(s1len + 1)]
    cdef list tmatrix = [[4 for i in range(s2len + 1)] for j in range(s1len + 1)]

    # compute scores and traceback matrix
    best_score = 0.0
    for i in range(1, s1len + 1):
        for j in range(1, s2len + 1):

            # match
            mscore = smatrix[i-1][j-1] + scores[i-1,j-1]

            # vertical
            if i == s1len:
                vscore = smatrix[i-1][j] + t_gap
            elif tmatrix[i-1][j] == VGAP:
                vscore = smatrix[i-1][j] + e_gap
            else:
                vscore = smatrix[i-1][j] + o_gap

            # horizontal
            if j == s2len:
                hscore = smatrix[i][j-1] + t_gap
            elif tmatrix[i][j-1] == HGAP:
                hscore = smatrix[i][j-1] + e_gap
            else:
                hscore = smatrix[i][j-1] + o_gap

            if mscore >= vscore and mscore >= hscore and mscore >= 0.0:
                smatrix[i][j] = mscore
                tmatrix[i][j] = MATCH
            elif vscore >= hscore and vscore >= 0.0:
                smatrix[i][j] = vscore
                tmatrix[i][j] = VGAP
            elif hscore >= 0.0:
                smatrix[i][j] = hscore
                tmatrix[i][j] = HGAP
            else:
                smatrix[i][j] = 0.0
                tmatrix[i][j] = END

            if smatrix[i][j] >= best_score:
                best_score = smatrix[i][j]
                row, col = i, j

    best_score = smatrix[row][col]

    if only_score:
        return best_score

    move = tmatrix[row][col]

    # traceback
    while move != END:
        if move == MATCH:
            a1.append(row - 1)
            a2.append(col - 1)
            row -= 1
            col -= 1
        elif move == VGAP:
            a1.append(row - 1)
            a2.append(GAP_SYM)
            row -= 1
        else:
            a1.append(GAP_SYM)
            a2.append(col - 1)
            col -= 1
        move = tmatrix[row][col]

    return a1[::-1], a2[::-1], best_score



import time
import gzip
import csv
import logging
import importlib.resources

import tqdm
import pandas as pd
import numpy as np
import scipy.sparse
from sklearn.metrics import pairwise_distances, pairwise_kernels

from retrieve.methods import pairwise_kernels_chunked
from retrieve.sparse_utils import top_k, substract_vector


logger = logging.getLogger(__name__)

SIM, DIST = 0, 1


def load_embeddings(path, vocab=None):
    # handle word2vec format
    skiprows = 0
    with open(path) as f:
        if len(next(f).strip().split()) == 2:
            skiprows = 1

    embs = pd.read_csv(
        path, sep=" ", header=None,
        index_col=0, skiprows=skiprows, quoting=csv.QUOTE_NONE)
    embs = embs.dropna(axis=1, how='all')
    embs = embs.T

    if vocab is not None:
        # drop words not in vocab
        missing = embs.columns.difference(vocab)
        logger.info("Dropping {} words from vocabulary".format(len(missing)))
        embs.drop(missing, 1, inplace=True)

    return embs


def normalize_vectors(vectors):
    return vectors / np.linalg.norm(vectors, axis=1)[:, None]


class Embeddings:
    # class constants
    SIM = SIM
    DIST = DIST

    """
    Convenience class to handle embeddings. This class is better initialized
    from the method `from_csv`

    Arguments
    =========
    keys : list of strings representing the words in the rows of vectors
    vectors : an np.array(n_words, dim_size)
    """
    def __init__(self, keys, vectors):
        if len(keys) != len(vectors):
            raise ValueError("Expected {} vectors".format(len(keys)))
        self.word2id = {}
        self.id2word = {}
        for idx, word in enumerate(keys):
            self.word2id[word] = idx
            self.id2word[idx] = word
        self.vectors = vectors

    def __len__(self):
        return len(self.word2id)

    def __getitem__(self, key):
        return self.vectors[self.word2id[key]]

    def __contains__(self, key):
        return key in self.word2id

    def apply_projection(self, proj, batch_size=4096, renorm=False):
        import torch
        if isinstance(proj, str):
            # assume path
            proj = torch.load(proj)
        vectors = self.vectors
        if renorm:
            vectors = normalize_vectors(vectors)
        for i in tqdm.tqdm(range(0, len(self.vectors), batch_size)):
            start, stop = i, min(i + batch_size, len(vectors))
            vectors[start:stop, :] = (proj @ vectors[start:stop, :].T).T
        self.vectors = vectors

    def normalize_vectors(self):
        self.vectors = normalize_vectors(self.vectors)

    @property
    def keys(self):
        return dict(self.word2id)

    def get_vectors(self, keys):
        targets = [w for w in keys if w in self.word2id]
        return targets, np.array(list(map(self.__getitem__, targets)))

    def default_vector(self):
        return np.mean(self.vectors, 0)

    @classmethod
    def from_fasttext(cls, path, vocab):
        if vocab is None:
            raise ValueError("FastText model requires vocab")
        try:
            import fastText
        except ModuleNotFoundError:
            raise ValueError("Couldn't import `fastText` module")
        model, vectors = fastText.load_model(path), []

        for word in vocab:
            vectors.append(model.get_word_vector(word))

        return cls(vocab, np.array(vectors))

    @classmethod
    def from_file(cls, path, vocab=None, skip_header=False):
        # dispatch fastText
        if path.endswith('bin'):
            return cls.from_fasttext(path, vocab)

        if vocab is not None:
            vocab = set(vocab)
            logger.info("Loading {} word embeddings".format(len(vocab)))
        keys, vectors = [], []

        open_fn = gzip.open if path.endswith(".gz") else open
        with open_fn(path) as f:
            if skip_header:
                next(f)
            for line in f:
                if isinstance(line, bytes):
                    line = line.decode()
                word, *vec = line.strip().split()
                if vocab and word not in vocab:
                    continue
                keys.append(word)
                vectors.append(np.array(vec, dtype=np.float))

        # report missing
        if vocab is not None:
            logger.info("Loaded {}/{} words from vocabulary".format(
                len(keys), len(vocab)))

        return cls(keys, np.array(vectors))

    @classmethod
    def from_csv(cls, path, vocab=None):
        """
        Arguments
        =========

        path : str, path to file with embeddings in csv format
            (word is assumed to go in first column)

        vocab : optional, subset of words to load

        Output
        ======

        keys : dict, mapping words to the index in indices respecting the
            order in which the keys appear
        indices : list, mapping keys to the index in embedding matrix
        """
        df = load_embeddings(path, vocab=vocab)
        return cls(list(df.keys()), np.array(df).T)

    @classmethod
    def from_resource(cls, path, vocab=None):
        if not importlib.resources.is_resource('retrieve.resources.misc', path):
            raise ValueError("Unknown resource: {}".format(path))
        with importlib.resources.path('retrieve.resources.misc', path) as f:
            return cls.from_file(str(f), vocab=vocab)

    def to_csv(self, path):
        with open(path, 'w') as f:
            for idx, word in sorted(self.id2word.items()):
                vec = ["{:.6}".format(i) for i in self.vectors[idx].tolist()]
                f.write(word + '\t' + ' '.join(vec) + '\n')

    def get_indices(self, words):
        keys, indices = {}, []
        for idx, w in enumerate(words):
            if w in self.word2id:
                keys[w] = idx
                indices.append(self.word2id[w])
        return keys, indices

    def get_S(self, vocab=None, fill_missing=False,
              metric='cosine', beta=1, apply_mod=True, cutoff=0.0, chunk_size=0):
        """
        Arguments
        =========

        vocab : list (optional), vocab in desired order. The output matrix will
            have word-similarities ordered according to the order in `vocab`.
            However, if `fill_missing` is False, while the order is mantained,
            there will be gaps.

        fill_missing : bool, whether to fill similarities with one-hot vectors
            for out-of-vocabulary words

        Output
        ======
        keys : list of words ordered as the output matrix
        S : np.array (or scipy.sparse.lil_matrix) (vocab x vocab), this will be
            a sparse array if a positive `cutoff` is passed

        >>> vectors = [[0.35, 0.75], [0.5, 0.5], [0.75, 0.35]]
        >>> embs = Embeddings(['a', 'c', 'e'], np.array(vectors))
        >>> vocab = ['c', 'd', 'a', 'f']
        >>> S = embs.get_S(vocab=vocab, fill_missing=True)
        >>> S.shape             # asked for 4 words (fill_missing)
        (4, 4)
        >>> S[1, 3] == 0.0      # missing words evaluate to one-hot vectors
        True
        >>> w1, w2 = embs['a'], embs['c']
        >>> sim = np.dot(w1, w2)/(np.linalg.norm(w1) * np.linalg.norm(w2))
        >>> np.allclose(S[0, 2], sim)
        True
        >>> S[2, 0] == S[0, 2]
        True
        >>> keys, S = embs.get_S(vocab=vocab)
        >>> list(keys) == ['c', 'a']  # words only in keys in requested order
        True
        >>> S.shape             # only words in space (fill_missing=False)
        (2, 2)
        >>> w1, w2 = embs['a'], embs['c']
        >>> sim = np.dot(w1, w2)/(np.linalg.norm(w1) * np.linalg.norm(w2))
        >>> np.allclose(S[0, 1], sim)
        True
        """
        if fill_missing and not vocab:
            raise ValueError("`fill_missing` requires `vocab`")
        if apply_mod and beta > 1 and cutoff is not None and cutoff < 0:
            raise ValueError("Negative cutoff with positive beta yields wrong results")

        keys, indices = self.get_indices(vocab or self.keys)
        if not keys:
            raise ValueError("Couldn't find any of the requested vocab")

        # (found words x found words)
        logger.info("Computing {} similarities".format(len(indices)))
        start = time.time()
        S = pairwise_kernels_chunked(
            self.vectors[indices], metric=metric, chunk_size=chunk_size,
            threshold=cutoff)
        logger.info("Got S in {:.2f} secs".format(time.time() - start))
        # apply modifications on S
        if apply_mod:
            S = (S.power(beta) if scipy.sparse.issparse(S) else np.power(S, beta))
        # add one-hot vectors for OOV and rearrange to match input vocabulary
        if fill_missing:
            # (requested words x requested words)
            S_ = scipy.sparse.lil_matrix((len(vocab), len(vocab)))
            # rearrange
            index = np.array([keys[w] for w in vocab if w in keys])
            index = np.tile(index, (len(index), 1))
            S_[index, index.T] = S
            S = S_
            # make sure diagonal is always 1
            S.setdiag(1)

            return S.tocsr()

        return keys, S

    def nearest_neighbours(self, words, n=10,
                           metric='cosine',
                           metric_type=SIM, csls_k=0):
        """
        If `metric_type` is Embeddings.SIM then `metric` must be one of
        sklearn.metrics.pairwise.PAIRWISE_KERNEL_FUNCTIONS:
        - ['cosine', 'sigmoid', 'linear', etc.]
        If `metric_type` is Embeddings.DIST then `metric` must be one of
        sklearn.metrics.pairwise.PAIRWISE_DISTANCE_FUNCTIONS:
        - ['cosine', 'euclidean', 'l1', 'l2', 'manhattan', 'cityblock']
        """
        if csls_k > 0 and metric_type != SIM:
            raise ValueError("CSLS is defined over similarities not distances")

        keys, index = self.get_indices(words)

        if metric_type == Embeddings.SIM:
            if csls_k > 0:
                S = csls(
                    pairwise_kernels(self.vectors, metric=metric, n_jobs=-1),
                    csls_k)
                S = S[index]
            else:
                S = pairwise_kernels(
                    self.vectors[index], self.vectors, metric=metric, n_jobs=-1)
            # get neighbours
            neighs = np.argsort(-S, axis=1)[:, 1: n+1]
        elif metric_type == Embeddings.DIST:
            S = pairwise_distances(
                self.vectors[index], self.vectors, metric=metric, n_jobs=-1)
            neighs = np.argsort(S, axis=1)[:, 1: n+1]
        else:
            raise ValueError("Unknown `metric_type`")

        S = S[np.arange(len(keys)).repeat(n), np.ravel(neighs)]
        S = S.reshape(len(keys), -1)
        # human form
        neighs = [{self.id2word[neighs[i, j]]: S[i, j] for j in range(n)}
                  for i in range(len(keys))]

        return keys, neighs


def train_gensim_embeddings(path, output_path=None, **kwargs):
    from gensim.models import Word2Vec
    from gensim.models.word2vec import LineSentence
    from retrieve import enable_log_level

    m = Word2Vec(sentences=LineSentence(path), **kwargs)
    if output_path:
        m.wv.save_word2vec_format(output_path)

    return m


def export_fasttext_embeddings(path, vocab, output_path=None):
    try:
        import fastText
    except ModuleNotFoundError:
        raise ValueError("Couldn't import `fastText` module")

    model = fastText.load(path)
    keys, vectors = {}, []
    for idx, word in enumerate(vocab):
        keys[word] = idx
        vectors.append(model.get_word_vector(word))

    if output_path is not None:
        with open(output_path, 'w+') as f:
            for word in keys:
                vec = ["{:.6}".format(i) for i in vectors[keys[word]].tolist()]
                f.write(word + '\t' + ' '.join(vec) + '\n')

    return keys, vectors


def csls_dense(S, k=10):
    indices, values = top_k(S, k + 1)
    mean_values = values[:, 1:].mean(1)
    return (2 * S) - mean_values[:, None] - mean_values[None, :]


def csls_sparse(S, k=10):
    indices, values = top_k(S, k + 1)
    mean_values = values[:, 1:].mean(1)
    S = substract_vector(
            substract_vector(S * 2, mean_values, axis=1),
                mean_values, axis=0)
    return S


def csls(S, k=10):
    if scipy.sparse.issparse(S):
        return csls_sparse(S, k)
    return csls_dense(S, k)


def csls_crosslingual(S, S1, S2, k=10):
    S1_mean = top_k(S1, k + 1)[1][:, 1:].mean(1)
    S2_mean = top_k(S2, k + 1)[1][:, 1:].mean(1)
    return (2 * S) - S1_mean[:, None] - S2_mean[None, :]


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('input')
    parser.add_argument('--output')
    parser.add_argument('--size', type=int, default=100)
    parser.add_argument('--window', type=int, default=5)
    parser.add_argument('--min_count', type=int, default=1)
    parser.add_argument('--workers', type=int, default=4)
    args = parser.parse_args()

    m = train_gensim_embeddings(args.input, output_path=args.output,
                                size=args.size, window=args.window,
                                min_count=args.min_count, workers=args.workers)



import multiprocessing as mp

import tqdm
import numpy as np
import scipy.sparse
from scipy.optimize import linear_sum_assignment

from .. import utils


def get_linear_sum_assignment_score(S, vocab1, vocab2, doc1, doc2, normalize=True):
    scores = np.zeros((len(doc1), len(doc2)))
    for idx, w1 in enumerate(doc1):
        for jdx, w2 in enumerate(doc2):
            scores[idx, jdx] = S[vocab1[w1], vocab2[w2] + len(vocab1)]

    a, b = linear_sum_assignment(scores, maximize=True)
    score = scores[a, b].sum()

    if normalize:
        score /= len(doc2)

    return score


class Workload:
    def __init__(self, S, vocab1, vocab2, coll1, coll2):
        self.S = S
        self.vocab1 = vocab1
        self.vocab2 = vocab2
        self.coll1 = coll1
        self.coll2 = coll2

    def __call__(self, chunk):
        output = []
        for i, j in chunk:
            doc1, doc2 = self.coll1[i].get_features(), self.coll2[j].get_features()
            score = get_linear_sum_assignment_score(
                self.S, self.vocab1, self.vocab2, self.coll1, self.coll2, self.normalize)
            output.append((i, j, score))
        return output


def parallel_search(S, vocab1, vocab2, coll1, coll2, items, processes, **kwargs):
    workload = Workload(S, vocab1, vocab2, coll1, coll2)
    rows, cols, data = [], [], []

    with mp.Pool(processes) as pool:
        chunks = list(utils.chunks(items, len(items) // (processes + 1)))
        for chunk in enumerate(pool.map(workload, chunks)):
            for i, j, score in chunk:
                rows.append(i)
                cols.append(j)
                data.append(score)

    return rows, cols, data


def parallel_linear_sum_assignment(
        coll1, coll2, S, vocab1, vocab2, pre_sims, normalize=True, processes=1):
    x, y, _ = scipy.sparse.find(pre_sims)
    
    processes = mp.cpu_count() if processes < 0 else processes
    if processes == 1:
        sims = scipy.sparse.dok_matrix((len(coll1), len(coll2)))
        for i, j in tqdm.tqdm(zip(x, y), total=len(x), desc='Linear Sum Assignment'):
            score = get_linear_sum_assignment_score(
                S, vocab1, vocab2, coll1[i].get_features(), coll2[j].get_features(),
                normalize=normalize)

    else:
        items = list(zip(x, y))
        rows, cols, data = parallel_search(
            S, vocab1, vocab2, coll1, coll2, items, processes, normalize=normalize)
        sims = scipy.sparse.csr_matrix(
            (data, (rows, cols)), shape=(len(coll1), len(coll2)))

    sims = sims.tocsr()
    sims.eliminate_zeros()

    return sims




import unittest

import random

from retrieve.methods import set_similarity
from retrieve.methods import SetSimilarity
from retrieve.corpora import load_vulgate


def _test_similarity_func(threshold, func, coll):
    return SetSimilarity(threshold, func).get_similarities(
        coll.get_features(cast=set, field='lemma'),
        coll.get_features(cast=set, field='lemma'),
        processes=5)


class TestSet(unittest.TestCase):
    def setUp(self):
        self.vulgate = load_vulgate(max_verses=1000)

    def _test_similarities(self, sims, n_samples, func):
        for (i, j), sim in random.sample(sims.todok().items(), n_samples):
            vul_a, vul_b = self.vulgate[i], self.vulgate[j]
            expected = getattr(vul_a, func)(vul_b, field='lemma')
            self.assertAlmostEqual(
                expected, sim,
                msg="{}:{}, {} expected: {}, but got {}".format(
                    i, j, func, expected, sim))

    def test_containment(self):
        self._test_similarities(
            _test_similarity_func(0.5, 'containment', self.vulgate),
            500, 'containment')

    def test_jaccard(self):
        self._test_similarities(
            _test_similarity_func(0.35, 'jaccard', self.vulgate),
            500, 'jaccard')

    def test_containment_min(self):
        self._test_similarities(
            _test_similarity_func(0.5, 'containment_min', self.vulgate),
            500, 'containment_min')

    def test_cosine(self):
        self._test_similarities(
            _test_similarity_func(0.5, 'cosine', self.vulgate),
            500, 'cosine')




import inspect

from .soft_cosine import soft_cosine_similarities, parallel_soft_cosine


def init_sklearn_vectorizer(vectorizer, vocab=None, **kwargs):
    params = set(inspect.signature(vectorizer).parameters)
    return vectorizer(
        vocabulary=list(vocab),
        # ensure l2 norm to speed up cosine similarity
        norm='l2',
        # overwrite default to avoid ignoring input
        token_pattern=r'\S+',
        **{k: v for k, v in kwargs.items() if k in params})


class VSM:
    def __init__(self, *args, **kwargs):
        """
        Dummy constructor, since not all methods require initalization
        """
        pass

    def fit(self, *args, **kwargs):
        """
        Dummy fit method (not compulsory)
        """
        return self

    def get_similarities(self, queries, index,
                         threshold=0.25, metric='linear',
                         chunk_size=500, disable_bar=False, **kwargs):
        """
        Get similarities according to specified kernel.

        Arguments
        =========

        threshold : float (optional), discard similarities below threshold,
            helps enforcing sparsity of the resulting matrix

        metric : str, (optional), default is "linear" which corresponds to
            a cosine similarity since all `retrieve` vectorizers output
            normalized vectors

        chunk_size : int, (optional), if greater than 0, the query matrix will
            be partitioned in chunk of `chunk_size` items and processed in batches.
            In combination with a certain `threshold`, this will allow to fit
            the whole (sparse) similarity matrix in memory
        """
        from retrieve.methods import pairwise_kernels_chunked

        index, queries = list(index), list(queries)
        transform = self.transform(queries + index)
        queries, index = transform[:len(queries)], transform[len(queries):]

        sims = pairwise_kernels_chunked(
            queries, index, metric=metric, chunk_size=chunk_size, threshold=threshold,
            desc='Chunked similarities', disable_bar=disable_bar)

        return sims


class VSMSoftCosine(VSM):
    def __init__(self, vocab, vectorizer, **kwargs):
        self.vectorizer = init_sklearn_vectorizer(vectorizer, vocab=vocab, **kwargs)

    def get_soft_cosine_similarities(self, queries, index, embs, threshold=0.25,
                                     chunk_size=500, disable_bar=False,
                                     parallel=False, n_jobs=-1, **kwargs):
        """
        Compute soft cosine similarities between queries and index using the
        (possibly sparse) similarity matrix S indexing the similarity between
        words i and j

        Arguments
        =========

        embs : Embeddings class that will be used to extract the similarity matrix
            for words in the inputs using the instantiated vectorizer

        chunk_size : int, (optional), if greater than 0, the query matrix will
            be partitioned in chunk of `chunk_size` items and processed in batches.
            In combination with a certain `threshold`, this will allow to fit
            the whole (sparse) similarity matrix in memory

        kwargs : extra arguments passed to embs.get_S
        """
        index, queries = list(index), list(queries)
        transform = self.transform(index + queries)
        index, queries = transform[:len(index)], transform[len(index):]
        S = embs.get_S(
            vocab=self.vectorizer.get_feature_names(), fill_missing=True, **kwargs)
        if parallel:
            sims = parallel_soft_cosine(
                queries, index, S, threshold=threshold, n_jobs=n_jobs)
        else:
            sims = soft_cosine_similarities(
                queries, index, S, chunk_size=chunk_size, threshold=threshold,
                disable_bar=disable_bar)

        return sims




import json
import contextlib
import collections

import numpy as np

from retrieve import sparse_utils


def _get_matches_fneg_tpos_at(q_idxs, i_idxs, ranking, at, strict):
    matches = []
    # [[i, i, i, ... (len(q_idxs))],
    #  [j, j, j, ...], ...
    #  [k, k, k, ...]]
    i_idxs_array = np.repeat(i_idxs[:, None], len(q_idxs), 1)
    # [i, j, k, i, j, k, ... len(q_idxs) * len(i_idxs)]
    i_idxs_array = np.ravel(i_idxs_array.T)
    # [[6, 4, 9, ... at],
    #   .
    #   .
    #   len(i_idxs)
    #  [3, 1, -1, -1, ... at],
    #   .
    #   .
    #   len(i_idxs)
    #  .
    #  .
    #  len(q_idxs) * len(i_idxs)
    q_idxs_array = np.repeat(ranking[q_idxs, :at], len(i_idxs), axis=0)

    row_matches, row_ranks = np.where(i_idxs_array[:, None] == q_idxs_array)
    # x-axis gives matches:
    # - rows of q_idxs (mod len i_idxs)
    # - rows of i_idxs (div len i_idxs)
    q_matches, i_matches = divmod(row_matches, len(i_idxs))
    # get original row and col indices
    q_matches, i_matches = q_idxs[q_matches], i_idxs[i_matches]
    # collect matches
    for q_match, i_match, rank in zip(q_matches.tolist(),
                                      i_matches.tolist(),
                                      row_ranks.tolist()):
        # rank should be 1-index
        matches.append((q_match, i_match, rank + 1))

    if strict:
        # how many were strictly retrieved
        tpos = len(q_matches)
        # how many of the many-to-many links were not retrieved
        fneg = len(q_idxs) * len(i_idxs) - len(q_matches)
    else:
        # at least one was retrieved
        # consider the whole match holistically
        tpos = int(len(matches) != 0)
        fneg = int(len(matches) == 0)

    return matches, tpos, fneg


def _get_fpos_at(unchecked, ranking, at):
    # empty ranks are represented as -1
    # considers a false positive if at least one indexed doc is retrieved within
    # the first ranked `at` documents
    fpos = np.sum(np.any(0 < ranking[unchecked, :at], axis=1))
    return fpos


def _get_tneg_at(ranking, refs, at):
    pred = np.sum(ranking[:, :at] == -1, axis=1) == at
    idxs = np.arange(len(ranking))
    true = np.array(list(set([i for tup, _  in refs for i in tup])))
    true = idxs[np.in1d(idxs, true, invert=True)]
    return np.sum(np.in1d(true, pred))


def ranking_stats_from_tuples(ranking, refs, at_values=(1, 5, 10, 20), strict=False):
    """
    Computes evaluation stats (true positives, false negatives and false positives)
    from a ranking matrix and an input set of true references.

    It considers a hit any link ranked within the first `at` items. Multiple
    cutoffs can be passed. The output is a list of dictionaries holding the results
    for the corresponding cutoff value.

    Arguments
    =========

    ranking : np.array(n, top_k), ranking array which has been pruned to the top k
        document hits
    refs : list[queries, indexed] where `queries` and `indexed` are tuples of doc_ids
        mapping to query or indexed documents that refer or are referred to by the
        other documents.
    at_values : tuple of cutoff points to consider for the evaluation. The intuition
        is that we can modify what we consider a hit based on how many of the top
        ranked items we are willing to check.
    strict : book, if False a reference is consider a true positive if at least one of
        the possibly many links in the reference is retrieved

    Output
    ======
    stats : list of dictionaries collecting the statistics, each dictionary is
        structured as follows: {'tpos': ..., 'fpos': ..., 'fneg': ..., 'tneg': ...}

    matches : list of tuples, where each tuple contains a match (q_id, i_d, rank)
        indicating which query doc was correctly retrieved as a reference to which
        indexed doc and at which rank

    >>> import numpy as np
    >>> n_queries, n_index = 5, 6
    >>> ranking = np.zeros((n_queries, n_index)) - 1
    >>> for i in range(n_queries):
    ...     for j in range(n_index):
    ...         if j < i:
    ...             ranking[i, j] = i + j
    >>> ranking
    array([[-1., -1., -1., -1., -1., -1.],
           [ 1., -1., -1., -1., -1., -1.],
           [ 2.,  3., -1., -1., -1., -1.],
           [ 3.,  4.,  5., -1., -1., -1.],
           [ 4.,  5.,  6.,  7., -1., -1.]])
    >>> refs = [([0], [1, 2]), ([2, 3], [3, 5, 6]), ([4], [8])]
    >>> stats, matches = ranking_stats_from_tuples(ranking, refs, at_values=[5])
    >>> len(stats)
    1
    >>> len(matches)
    3
    >>> matches
    [(2, 3, 2), (3, 3, 1), (3, 5, 3)]
    >>> stats = stats[0]
    >>> stats['tpos'], stats['fneg'], stats['fpos'], stats['tneg']
    (1, 2, 2, 1)
    >>> stats, matches = ranking_stats_from_tuples(
    ...     ranking, refs, at_values=[5], strict=True)
    >>> stats = stats[0]
    >>> stats['tpos'], stats['fneg'], stats['fpos'], stats['tneg']
    (3, 6, 2, 1)
    """
    # >>> refs = [([0], [1, 2]), ([2, 3], [3, 5, 6]), ([4], [8])]
    # >>> array([[-1,  -1,  -1,  -1,  -1, -1],  # fneg
    #            [ 1., -1,  -1,  -1,  -1, -1],  # fpos
    #            [ 2.,  3., -1,  -1,  -1, -1],  # tpos
    #            [ 3.,  4.,  5., -1,  -1, -1],  # (tpos already checked)
    #            [ 4.,  5.,  6.,  7., -1, -1]]) # fneg & fpos
    matches = []
    stats = [collections.Counter() for _ in at_values]
    checked = collections.defaultdict(list)

    for q_idxs, i_idxs in refs:
        q_idxs, i_idxs = np.array(q_idxs), np.array(i_idxs)
        for idx_at, at in enumerate(sorted(at_values)):
            row_matches, tpos, fneg = _get_matches_fneg_tpos_at(
                q_idxs, i_idxs, ranking, at, strict)
            stats[idx_at]['tpos'] += tpos
            stats[idx_at]['fneg'] += fneg

            # collect rank data only at the highest at
            # the rank can be used to filter out by cutoff points later
            if at == max(at_values):
                matches.extend(row_matches)

            if row_matches:
                row_checked, _, _ = zip(*row_matches)
                checked[at].extend(row_checked)

    # false positives and true negatives
    for idx, at in enumerate(sorted(at_values)):
        unchecked = np.arange(len(ranking))
        unchecked = unchecked[np.isin(unchecked, np.array(checked[at]), invert=True)]
        stats[idx]['fpos'] += _get_fpos_at(unchecked, ranking, at)
        stats[idx]['tneg'] += _get_tneg_at(ranking, refs, at)

    return stats, matches


def f_score(p, r, beta=1):
    """
    F-score
    """
    return (1 + beta ** 2) * ((p * r) / (((beta ** 2) * p) + r))


def get_metrics_from_tpos_fneg_fpos(tpos, fneg, fpos):
    p = tpos / (tpos + fpos)
    r = tpos / (tpos + fneg)
    return {'p': p,
            'r': r,
            'f1': f_score(p, r, beta=1),
            'f2': f_score(p, r, beta=2),
            'fp5': f_score(p, r, beta=0.5)}


def get_metrics(inp, refs, at_values=(1, 5, 10, 20), strict=False, input_type='sims'):
    """
    Get metrics from similarity metric and refs. Input might be the full similarity
    matrix or an already computed top-k ranking
    """
    assert inp.shape[1] >= max(at_values), \
        "Asked to rank top-{} items but got only {} items".format(
            max(at_values), inp.shape[1])

    ranking = None
    if input_type == 'sims':
        ranking, _ = sparse_utils.top_k(inp, max(at_values))
    else:
        ranking = inp[:, :max(at_values)]

    stats, _ = ranking_stats_from_tuples(
        ranking, refs, at_values=at_values, strict=strict)

    results = {}
    for at, stats in zip(at_values, stats):
        tpos, fneg, fpos = stats['tpos'], stats['fneg'], stats['fpos']
        # accumulate
        for metric, score in get_metrics_from_tpos_fneg_fpos(tpos, fneg, fpos).items():
            results['{}@{}'.format(metric, at)] = score

    return results


def get_thresholded_metrics_approx(sims, refs, thresholds, copy=False, **kwargs):
    # drop items that could never make it to the top k (add a safety margin)
    top_k = max(kwargs.get('at_values', [20]))
    max_items = min(sims.nnz, sims.shape[0] * top_k * 5)
    perc = 1 - (max_items / len(sims.data))
    min_th = np.percentile(sims.data, 100 * perc, interpolation='lower')
    sims = sparse_utils.set_threshold(sims, min_th, copy=True)
    assert len(sims.data) >= max_items

    return get_thresholded_metrics(sims, refs, thresholds, copy=copy, **kwargs)


def get_thresholded_metrics(sims, refs, thresholds, copy=False, **kwargs):
    thresholds = sorted(thresholds)
    metrics = collections.defaultdict(list)

    # sort sims
    ranking, values = sparse_utils.top_k(sims, max(kwargs.get('at_values', [20])))
    # overwrite input type
    kwargs['input_type'] = 'ranking'

    for th in thresholds:
        x, y = np.where(values < th)
        ranking[x, y] = -1
        for metric, val in get_metrics(ranking, refs, **kwargs).items():
            metrics[metric].append(val)

    return metrics





@contextlib.contextmanager
def results_writer(fp):
    """
    Write output of experiment to file. For a given similarity matrix and refs
    compute evaluation results at given thresholds. It also registers the sparsity
    of the passed matrix (nnz) *before* applying a cutoff at a top-k value.
    If copy==True, do not modify insplace the similarity matrix
    """
    fp = open(fp, 'w+')

    def write(sims, refs, thresholds, copy=False, **kwargs):
        row = {}
        row['results'] = []
        row['nnz'] = []

        metrics = get_thresholded_metrics(sims, refs, thresholds, copy=copy)

        for th, nnz in zip(thresholds, metrics.pop('nnz')):
            row['nnz'].append({'th': float(th), 'value': nnz})
        for metric, vals in metrics.items():
            for th, val in zip(thresholds, vals):
                row['results'].append({'metric': metric, 'th': float(th), 'val': val})
        for key, val in kwargs.items():
            row[key] = val

        fp.write(json.dumps(row) + '\n')
        fp.flush()

    yield write

    fp.close()




import numpy as np
import scipy.sparse
import numba as nb


@nb.njit()
def substract_vector_loop(data, indices, indptr, vec):
    for i in nb.prange(vec.shape[0]):
        data[indptr[i]:indptr[i+1]] -= vec[i]
    return data, indices, indptr


def substract_vector(S, vec, inplace=False, axis=0):
    orig_type = type(S)
    X = S
    if not inplace:
        X = S.copy()
    if axis == 0:
        X = X.tocsr()
    elif axis == 1:
        X = X.tocsc()
    substract_vector_loop(X.data, X.indices, X.indptr, vec)
    return orig_type(X)


@nb.njit()
def _top_k_dense(data, indices, indptr, k):
    # indptr holds pointers to indices and data
    # indices[indptr[0]:indptr[1]] -> index of nonzero items in 1st row
    # data[indptr[0]:indptr[1]]    -> nonzero items in 1st row
    nrows = indptr.shape[0] - 1  # substract one because of format

    # output variables
    top_indices = np.zeros((nrows, k), dtype=indices.dtype) - 1
    top_vals = np.zeros((nrows, k), dtype=data.dtype) * np.nan

    for i in nb.prange(nrows):
        start, stop = indptr[i], indptr[i + 1]
        top_k = np.argsort(data[start:stop])[::-1][:k]
        n_items = min(len(top_k), k)
        # assign
        top_indices[i, 0:n_items] = indices[start:stop][top_k]
        top_vals[i, 0:n_items] = data[start:stop][top_k]

    return top_indices, top_vals


@nb.njit()
def _top_k_sparse_data(data, indices, indptr, k):
    # indptr holds pointers to indices and data
    # indices[indptr[0]:indptr[1]] -> index of nonzero items in 1st row
    # data[indptr[0]:indptr[1]]    -> nonzero items in 1st row
    nrows = indptr.shape[0] - 1  # substract one because of format

    top_indices = []
    top_vals = []
    top_rows = []
    top_cols = []

    for i in nb.prange(nrows):
        start, stop = indptr[i], indptr[i + 1]
        top_k = np.argsort(data[start:stop])[::-1][:k]
        if len(top_k) > 0:
            top_indices.append(indices[start:stop][top_k])
            top_vals.append(data[start:stop][top_k])
            top_cols.append(np.arange(0, len(top_k)))
            top_rows.append(np.repeat([i], len(top_k)))

    return top_indices, top_vals, top_rows, top_cols


def _top_k_sparse(data, indices, indptr, k):
    top_indices, top_vals, rows, cols = _top_k_sparse_data(data, indices, indptr, k)
    top_indices, top_vals = np.concatenate(top_indices), np.concatenate(top_vals)
    rows, cols = np.concatenate(rows), np.concatenate(cols)
    top_indices = scipy.sparse.csr_matrix(
        (top_indices, (rows, cols)), dtype=indices.dtype)
    top_vals = scipy.sparse.csr_matrix((top_vals, (rows, cols)), dtype=data.dtype)

    return top_indices, top_vals


def top_k(X, k):
    """
    X : matrix, (n x m)

    Output
    ======

    np.array(n x k), top_k items per row, it doesn't exclude the
        highest one, which is in self-search typically corresponds
        to itself

    >>> # checkerboard pattern with rowise increments
    >>> nrow = 3
    >>> ncol = 5
    >>> X = scipy.sparse.dok_matrix((nrow, ncol))
    >>> for i in range(nrow):
    ...     for j in range(i % 2, ncol, 2):
    ...         X[i, j] = X.nnz + 1
    >>> indices, data = top_k(X, 2)
    >>> indices.tolist()
    [[4, 2], [3, 1], [4, 2]]
    >>> data.tolist()
    [[3.0, 2.0], [5.0, 4.0], [8.0, 7.0]]
    """
    if scipy.sparse.issparse(X):
        X = X.tocsr()
        data, indices, indptr = X.data, X.indices, X.indptr
        top_indices, top_vals = _top_k_dense(data, indices, indptr, k)

    else:
        top_indices = np.argsort(X, 1)[:, -k:][:, ::-1]
        top_vals = np.take_along_axis(X, top_indices, 1)

    return top_indices, top_vals


def sparse_chunks(M, chunk_size):
    """
    This creates copy since sparse matrices don't have views
    """
    n, _ = M.shape
    for i in range(0, n, chunk_size):
        start, stop = i, min(i + chunk_size, n)
        yield (start, stop), M[start:stop]
    if stop < n:
        yield (stop, n), M[stop:n]


def set_threshold(X, threshold, sparse_matrix=scipy.sparse.csr_matrix, copy=False):
    """
    Threshold a given (possibly sparse matrix). This function
    will increase the sparsity of the matrix. If the input is not sparse
    it will default to numpy functionality.

    Arguments
    =========
    X : sparse_matrix or np.array
    threshold : float
    sparse_matrix : output sparse matrix type
    copy : whether to operate in place (only for sparse input)

    >>> import scipy.sparse
    >>> X = scipy.sparse.dok_matrix((10, 10))
    >>> X[0, 0] = 0.5
    >>> X[2, 4] = 0.25
    >>> X[7, 1] = 0.75
    >>> X[8, 7] = 0.15
    >>> X = X.tocsr()
    >>> set_threshold(X, 0.1, copy=True).nnz
    4
    >>> set_threshold(X, 0.3, copy=True).nnz
    2
    >>> set_threshold(X, 0.8, copy=True).nnz
    0
    >>> _ = set_threshold(X, 0.5)
    >>> X.nnz
    2
    >>> X_orig = X.copy()
    >>> _ = set_threshold(X, 0.6, copy=True)
    >>> (X != X_orig).nnz
    0
    >>> _ = set_threshold(X, 0.6, copy=False)
    >>> (X != X_orig).nnz > 0
    True
    """
    if threshold == 0:
        return X

    if not scipy.sparse.issparse(X):
        if copy:
            X = np.copy(X)
        X[np.where(X < threshold)] = 0.0
        return X

    if copy:
        rows, cols, _ = scipy.sparse.find(X >= threshold)
        if len(rows) > 0:
            # matrix is not empty
            data = np.squeeze(np.array(X[rows, cols]), axis=0)  # find returns matrix
            return sparse_matrix(
                scipy.sparse.csr_matrix(
                    (data, (rows, cols)), shape=X.shape))
        else:
            # matrix is empty
            return sparse_matrix(X.shape)


    if isinstance(X, (scipy.sparse.lil_matrix, scipy.sparse.dok_matrix)):
        raise ValueError("Cannot efficiently drop items on", str(type(X)))

    X.data[np.abs(X.data) < threshold] = 0.0
    X.eliminate_zeros()

    return X




import os

import scipy.sparse
import numpy as np

import flask
from flask import Flask
from flask import render_template


def extract_matching_words(doc1, doc2):
    doc1ws = [w for feat in doc1.get_features() for w in feat.split('--')]
    doc2ws = [w for feat in doc2.get_features() for w in feat.split('--')]
    intersect = set(doc1ws).intersection(doc2ws)
    doc1ids, doc2ids = [], []
    for idx, lem in enumerate(doc1.fields['lemma']):
        if lem in intersect:
            doc1ids.append(idx)
    for idx, lem in enumerate(doc2.fields['lemma']):
        if lem in intersect:
            doc2ids.append(idx)

    return doc1ids, doc2ids


class VisualizerApp:
    def __init__(self, sims, coll1, coll2=None, 
                 sim_range=(None, None), max_points=5000, sample=False,
                 host='localhost', port=5000):
        """
        Dot-plot visualization app
        """
        # data
        if coll2 is None:
            # drop diagonal
            sims = scipy.sparse.tril(sims)

        coll2 = coll2 or coll1
        # force long collection on y axis
        if len(coll1) > len(coll2):
            self.coll2 = coll1
            self.coll1 = coll2
            self.sims = sims.T
        else:
            self.coll1 = coll1
            self.coll2 = coll2
            self.sims = sims

        self.min_sim, self.max_sim = sim_range
        self.max_points = max_points
        self.sample = sample

        # app
        self.host = host
        self.port = port
        self.app = Flask(
            __name__,
            # this breaks if the app is moved to a different directory
            template_folder=os.path.join(
                os.path.dirname(os.path.abspath(__file__)), 'templates'))

        # add rules
        self.app.add_url_rule("/", "index", self.index)
        self.app.add_url_rule("/matching", "matching", self.matching, methods=['GET'])
        self.app.add_url_rule("/heatmap", "heatmap", self.heatmap, methods=['GET'])

    def index(self):
        return render_template("index.html")

    def matching(self, ctx=2):
        data = flask.request.args
        row, col = int(data['row']), int(data['col'])
        doc1, doc2 = self.coll1[row], self.coll2[col]
        doc1ids, doc2ids = extract_matching_words(doc1, doc2)
        # context doc1
        doc1l = [self.coll1[i] for i in range(max(0, row-ctx), row)]
        doc1r = [self.coll1[i] for i in range(row+1, min(len(self.coll1), row+ctx+1))]
        # context doc2
        doc2l = [self.coll2[i] for i in range(max(0, col-ctx), col)]
        doc2r = [self.coll2[i] for i in range(col+1, min(len(self.coll2), col+ctx+1))]

        return {'doc1': {'left': ' '.join(d.text for d in doc1l),
                         'right': ' '.join(d.text for d in doc1r),
                         'text': doc1.text, 'match': doc1ids,
                         'id': doc1.get_printable_doc_id()},
                'doc2': {'left': ' '.join(d.text for d in doc2l),
                         'right': ' '.join(d.text for d in doc2r),
                         'text': doc2.text, 'match': doc2ids,
                         'id': doc2.get_printable_doc_id()}}

    def heatmap(self):
        rows, cols, vals = scipy.sparse.find(self.sims)
        n_points, sampled = len(vals), False
        if len(vals) >= self.max_points:
            sampled = True
            if self.sample:
                index = np.random.choice(
                    np.arange(len(vals)), size=self.max_points, replace=False)
            else:
                index = np.argsort(vals)[-self.max_points:]
            rows, cols, vals = rows[index], cols[index], vals[index]

        matches = list(zip(rows.tolist(), cols.tolist(), vals.tolist()))
        min_sim = float(vals.min()) if self.min_sim is None else self.min_sim
        max_sim = float(vals.max()) if self.max_sim is None else self.max_sim
        data = {'points': [{'row': row,
                            'col': col,
                            'row_id': self.coll1[row].get_printable_doc_id(),
                            'col_id': self.coll2[col].get_printable_doc_id(),
                            'sim': val} for row, col, val in matches],
                'nrow': len(self.coll1),
                'ncol': len(self.coll2),
                'rowName': self.coll1.name,
                'colName': self.coll2.name,
                'meanSim': float(vals.mean()),
                'maxSim': max_sim,
                'minSim': min_sim,
                'sampled': sampled,
                'nPoints': n_points}

        return data

    def run(self, debug=True):
        self.app.run(host=self.host, port=self.port, debug=debug)




import os
import logging
import time
import contextlib
import itertools
import unicodedata
import sys
if sys.version_info.minor < 7:
    from importlib_resources import open_text, is_resource
else:
    from importlib.resources import open_text, is_resource


logger = logging.getLogger(__name__)


class Ngrams:
    def __init__(self, min_n=1, max_n=1, skip_k=0, sep='--'):
        self.min_n = min_n
        self.max_n = max_n
        self.skip_k = skip_k
        self.sep = sep

    def get_summary(self):
        return (self.min_n, self.max_n, self.skip_k)

    def get_ngrams_generator(self, s):
        for ngram in get_ngrams(
                s,
                min_n=self.min_n,
                max_n=self.max_n,
                skip_k=self.skip_k):
            yield self.sep.join(ngram)

    def get_ngrams(self, s):
        return list(self.get_ngrams_generator(s))


def get_ngrams(s, min_n=1, max_n=1, skip_k=0):
    """
    N-gram generator over input sequence. Allows multiple n-gram orders at once
    as well as skip-grams
    """
    for n in range(min_n, max_n + 1):
        if skip_k and n > 1:
            for ngram in zip(*[s[i:] for i in range(n + skip_k)]):
                first, rest = ngram[:1], ngram[1:]
                for tail in itertools.combinations(rest, n - 1):
                    yield first + tail
        else:
            for ngram in zip(*[s[i:] for i in range(n)]):
                yield ngram


def chunks(it, size):
    """
    Chunk a generator into a given size (last chunk might be smaller)
    """
    buf = []
    for s in it:
        buf.append(s)
        if len(buf) == size:
            yield buf
            buf = []
    if len(buf) > 0:
        yield buf


def drop_string_diacritics(s):
    return ''.join(c for c in unicodedata.normalize('NFD', s)
                   if unicodedata.category(c) != 'Mn')


class Stopwords:
    def __init__(self, *paths, drop_diacritics=False):
        self.words = set()
        for path in paths:
            self.words.update(load_stopwords(path, drop_diacritics))
        self.paths = paths
        self.drop_diacritics = drop_diacritics

    def get_summary(self):
        return {'paths': self.paths, 'drop_diacritics': self.drop_diacritics}

    def __contains__(self, w):
        return w in self.words


def load_stopwords(path, drop_diacritics=False):
    """
    Load stopwords from vertical format file. Ignore comments (#) and (?) doubted
    """
    if os.path.isfile(path):
        with open(path) as f:
            lines = list(f)
    elif is_resource('retrieve.resources.stop', path):
        # try with package resource
        lines = list(open_text('retrieve.resources.stop', path))
    else:
        raise ValueError("Couldn't find file: '{}'".format(path))

    stopwords = []
    for line in lines:
        line = line.strip()
        if line.startswith('?') or line.startswith('#'):
            continue
        if not line:
            continue
        if drop_diacritics:
            line = drop_string_diacritics(line)
        stopwords.append(line)

    return set(stopwords)


def load_freqs(path, top_k=0):
    """
    Load frequencies from file in format:

        word1 123
        word2 5
    """
    freqs = {}
    with open(path) as f:
        for line in f:
            count, w = line.strip().split()
            freqs[w] = int(count)
            if top_k > 0 and len(freqs) >= top_k:
                break
    return freqs


@contextlib.contextmanager
def timer(print_on_leave=True, desc='', fmt='took {:0.1f} secs in total', **kwargs):

    start = last = time.time()

    def time_so_far(desc='', fmt='took {:0.1f} secs', **kwargs):
        nonlocal last
        desc = ' - ' + (desc + ' ' if desc else desc)
        took = time.time() - last
        logger.info(desc + fmt.format(took), **kwargs)
        last = time.time()
        return took

    yield time_so_far

    desc = desc + ' ' if desc else desc

    logger.info(desc + fmt.format(time.time() - start), **kwargs)




from .align import *
from .set_similarity import *
from .vsm import *

try:
    import pyemd
    from .wmd import get_wmd, get_wmd_histograms
except ModuleNotFoundError:
    import warnings
    warnings.warn("Couldn't import `pyemd` module")
    
from .pairwise_chunked import pairwise_kernels_chunked
from .transportation import parallel_linear_sum_assignment
from .transportation import get_linear_sum_assignment_score




import tqdm
import scipy.sparse
from sklearn.metrics.pairwise import pairwise_kernels

from ..sparse_utils import sparse_chunks, set_threshold


def pairwise_kernels_chunked(X, Y=None, metric='linear', chunk_size=0,
                             desc=None, disable_bar=True, threshold=None, n_jobs=-1):
    """
    Chunked version of pairwise_kernels that applies thresholds to produce
    sparse similarity matrices
    """
    if Y is None:
        Y = X

    if chunk_size > 0:
        (n, _), (m, _) = X.shape, Y.shape
        sims = scipy.sparse.lil_matrix((n, m))
        n_chunks = n // chunk_size
        for (i_start, i_stop), Q in tqdm.tqdm(sparse_chunks(X, chunk_size),
                                              total=n_chunks, desc=desc,
                                              disable=disable_bar):
            Q_sims = pairwise_kernels(Q, Y, metric=metric, n_jobs=-1)
            if threshold is not None:
                set_threshold(Q_sims, threshold)
                sims[i_start: i_stop, :] = Q_sims
    else:
        sims = pairwise_kernels(X, Y, metric=metric, n_jobs=n_jobs)
        if threshold is not None:
            set_threshold(sims, threshold)

    if scipy.sparse.isspmatrix_lil(sims):
        sims = sims.tocsr()

    return sims




import math
import logging

import numpy as np
import scipy.sparse
from joblib import Parallel, delayed, effective_n_jobs

import tqdm

from ...sparse_utils import sparse_chunks, set_threshold


logger = logging.getLogger(__name__)


def soft_cosine_simple(query, index, S):
    if scipy.sparse.issparse(query):
        query = dict(query.todok())
    if scipy.sparse.issparse(index):
        index = dict(index.todok())

    num = den1 = den2 = 0
    for _, i in set(query).union(index):
        for _, j in set(query).union(index):
            sim = S[i, j]
            num  += sim * query.get((0, i), 0) * index.get((0, j), 0)
            den1 += sim * query.get((0, i), 0) * query.get((0, j), 0)
            den2 += sim * index.get((0, i), 0) * index.get((0, j), 0)

    if den1 == 0 or den2 == 0:
        return 0.0

    return (num / (math.sqrt(den1) * math.sqrt(den2)))


def soft_cosine_similarities(queries, index, S, chunk_size=500, threshold=0.0,
                             disable_bar=False):
    """
    This function assumes that the order of vocabulary in the similarity matrix
    correspondes to the order of the vocabulary in the document representations

    Arguments
    =========

    queries : np.array(n, vocab), n query docs in BOW format, possibly sparse
    index : np.array(m, vocab), m indexed docs in BOW format, possibly sparse
    S : np.array(vocab, vocab), similarity matrix (possibly raised to a power)
        possibly a sparse matrix
    chunk_size : int, specifies the number of rows in queries to be computed
        at once. Helpful for handling memory constraints
    threshold : float, sparsify any output cells below such threshold

    Output
    ======
    sims : np.array(n, m), soft cosine similarities
    """
    is_S_sparse = scipy.sparse.issparse(S)

    (n, _), (m, _) = queries.shape, index.shape
    sims = scipy.sparse.lil_matrix((n, m)) if is_S_sparse else np.zeros((n, m))

    # (vocab x m) assumes m << n, so this can be stored in memory
    SindexT = S @ index.T
    # (m x vocab) x (m x vocab) -> (m x vocab)
    den2 = index.multiply(SindexT.T)
    # (m x vocab) -> (m)
    den2 = np.squeeze(np.array(den2.sum(1)))
    den2 = np.sqrt(den2)

    n_chunks = queries.shape[0] // chunk_size
    for (i_start, i_stop), Q in tqdm.tqdm(sparse_chunks(queries, chunk_size),
                                          total=n_chunks, desc='Soft cosine',
                                          disable=disable_bar):
        # (chunk_size x vocab) @ (vocab x m) -> (chunk_size x m)
        num = Q @ SindexT

        # (vocab x vocab) x (vocab x chunk_size) -> (vocab x chunk_size)
        SqueryT = S @ Q.T
        # (chunk_size x vocab) x (chunk_size x vocab) -> (chunk_size x vocab)
        den1 = Q.multiply(SqueryT.T)
        # (chunk_size x vocab) -> (chunk_size)
        den1 = np.squeeze(np.array(den1.sum(1)))
        den1 = np.sqrt(den1)

        # (chunk_size x m)
        Q_sims = (den1[:, None] * den2[None, :])
        with np.errstate(divide='ignore'):
            # this might run into division by 0
            Q_sims = num.multiply(1 / Q_sims) if is_S_sparse else (num / Q_sims)

        Q_sims = np.nan_to_num(Q_sims, copy=False)

        # apply threshold
        set_threshold(Q_sims, threshold)

        sims[i_start:i_stop, :] = Q_sims

    if scipy.sparse.isspmatrix_lil(sims):
        sims = sims.tocsr()

    return sims


def _parallel_soft_cosine1(index, S):
    # (vocab x m) assumes m << n, so this can be stored in memory
    SindexT = S @ index.T
    # (m x vocab) x (m x vocab) -> (m x vocab)
    den2 = index.multiply(SindexT.T)
    # (m x vocab) -> (m)
    den2 = np.squeeze(np.array(den2.sum(1)))
    den2 = np.sqrt(den2)

    return SindexT, den2


def _parallel_soft_cosine2(Q, SindexT, S, den2, is_S_sparse):
    # (chunk_size x vocab) @ (vocab x m) -> (chunk_size x m)
    num = Q @ SindexT

    # (vocab x vocab) x (vocab x chunk_size) -> (vocab x chunk_size)
    SqueryT = S @ Q.T
    # (chunk_size x vocab) x (chunk_size x vocab) -> (chunk_size x vocab)
    den1 = Q.multiply(SqueryT.T)
    # (chunk_size x vocab) -> (chunk_size)
    den1 = np.squeeze(np.array(den1.sum(1)))
    den1 = np.sqrt(den1)

    # (chunk_size x m)
    Q_sims = (den1[:, None] * den2[None, :])
    with np.errstate(divide='ignore'):
        # this might run into division by 0
        Q_sims = num.multiply(1 / Q_sims) if is_S_sparse else (num / Q_sims)

    Q_sims = np.nan_to_num(Q_sims, copy=False)

    return Q_sims


def wrapper_fn_assign(sims, i_start, i_stop, threshold, fn, *args, **kwargs):
    out = fn(*args, **kwargs)
    set_threshold(out, threshold)
    sims[i_start:i_stop] = out


def get_chunks(M, n_chunks):
    n, _ = M.shape
    chunk_size = n // n_chunks
    for i in range(n_chunks):
        start = i * chunk_size
        stop = min((i + 1) * chunk_size, n)
        yield start, stop
    if stop < n:
        yield stop, n


def parallel_soft_cosine(queries, index, S, threshold=0.0, n_jobs=-1):
    """
    Same as soft_cosine_similarities but with parallelized computations

    Arguments
    =========

    queries : np.array(n, vocab), n query docs in BOW format, possibly sparse
    index : np.array(m, vocab), m indexed docs in BOW format, possibly sparse
    S : np.array(vocab, vocab), similarity matrix (possibly raised to a power),
        possibly a sparse matrix
    threshold : float, sparsify any output cells below such threshold
    n_jobs : int, number of jobs to run in parallel

    Output
    ======
    sims : np.array(n, m), soft cosine similarities
    """
    is_S_sparse = scipy.sparse.issparse(S)

    (n, _), (m, _) = queries.shape, index.shape

    SindexT, den2 = _parallel_soft_cosine1(index, S)

    # following sklearn/metrics/pairwise
    fd = delayed(wrapper_fn_assign)
    sims = scipy.sparse.lil_matrix((n, m)) if is_S_sparse else np.zeros((n, m))

    n_chunks = effective_n_jobs(n_jobs)
    logger.info("Using {} CPUs".format(n_chunks))
    Parallel(backend='threading', n_jobs=n_chunks)(
        fd(sims, i_start, i_stop, threshold,
           # fn
           _parallel_soft_cosine2,
           # args
           queries[i_start: i_stop], SindexT, S, den2, is_S_sparse)
        for i_start, i_stop in get_chunks(queries, n_chunks))

    if scipy.sparse.isspmatrix_lil(sims):
        sims = sims.tocsr()

    return sims


if __name__ == '__main__':
    import timeit

    from retrieve.corpora import load_vulgate
    from retrieve.data import Criterion, TextPreprocessor, FeatureSelector
    from retrieve import utils
    from retrieve.embeddings import Embeddings
    from retrieve.methods import Tfidf

    # load
    vulg = load_vulgate(include_blb=True)
    # preprocess
    TextPreprocessor(
        stopwords=utils.Stopwords('data/stop/latin_berra.stop')
    ).process_collections(vulg, min_n=1, max_n=2, sep=' ')
    # drop features and get vocabulary
    fsel = FeatureSelector(vulg)
    vocab = fsel.filter_collections(
        vulg, criterion=(Criterion.DF >= 2) & (Criterion.FREQ >= 5))
    # get documents
    feats = vulg.get_features()
    # transform to tfidf
    tfidf = Tfidf(vocab).fit(feats)
    feats = tfidf.transform(feats)
    query, index = feats[:10000], feats[10000:20000]
    vocab = [ft for ft in tfidf.vectorizer.get_feature_names() if len(ft.split()) == 1]
    # load embeddings, make sure S is in same order as vocab
    embs = Embeddings.from_resource(
        'latin.lemma.ft.dim100.mc2.embeddings.gz', vocab=vocab)

    for cutoff in [0.25, 0.5, 0.7, 0.8, 0.85, 0.9, 0.95]:
        S = embs.get_S(
            vocab=tfidf.vectorizer.get_feature_names(),
            fill_missing=True, cutoff=cutoff)
        nelem = S.shape[0] ** 2
        total_sim_size = (nelem - S.shape[0]) / 2
        effec_sim_size = (S.nnz - S.shape[0]) / 2
        print(" - Sparse run for threshold {:.2f}".format(cutoff))
        print("     - Off-diagonal pairs {}: {:.2f}%".format(
            int(effec_sim_size), effec_sim_size / total_sim_size))

        a = timeit.Timer(lambda: soft_cosine_similarities(
            query, index, S, threshold=0.25, chunk_size=100, disable_bar=True)
        ).timeit(number=5)
        b = timeit.Timer(lambda: soft_cosine_similarities(
            query, index, S.todense(), threshold=0.25, chunk_size=100, disable_bar=True)
        ).timeit(number=5)
        print("   - Runtime sparse: {:.3f}".format(a))
        print("   - Runtime dense: {:.3f}".format(b))

        a = timeit.Timer(lambda: parallel_soft_cosine(
            query, index, S, threshold=0.25)
        ).timeit(number=5)
        b = timeit.Timer(lambda: parallel_soft_cosine(
            query, index, S.todense(), threshold=0.25)
        ).timeit(number=5)
        print("   - Runtime parallel sparse: {:.3f}".format(a))
        print("   - Runtime parallel dense: {:.3f}".format(b))

    # - Sparse run for threshold 0.25
    #     - Off-diagonal pairs 4238493: 0.33%
    #   - Runtime sparse: 40.134
    #   - Runtime dense: 3.659
    # - Sparse run for threshold 0.50
    #     - Off-diagonal pairs 264663: 0.02%
    #   - Runtime sparse: 7.002
    #   - Runtime dense: 3.718
    # - Sparse run for threshold 0.70
    #     - Off-diagonal pairs 4877: 0.00%
    #   - Runtime sparse: 2.678
    #   - Runtime dense: 3.770
    # - Sparse run for threshold 0.80
    #     - Off-diagonal pairs 556: 0.00%
    #   - Runtime sparse: 2.666
    #   - Runtime dense: 3.621
    # - Sparse run for threshold 0.85
    #     - Off-diagonal pairs 209: 0.00%
    #   - Runtime sparse: 2.658
    #   - Runtime dense: 3.719
    # - Sparse run for threshold 0.90
    #     - Off-diagonal pairs 65: 0.00%
    #   - Runtime sparse: 2.507
    #   - Runtime dense: 3.672
    # - Sparse run for threshold 0.95
    #     - Off-diagonal pairs 2: 0.00%
    #   - Runtime sparse: 2.553
    #   - Runtime dense: 3.604




import math
import logging
import multiprocessing as mp

from scipy.sparse import dok_matrix, csr_matrix
from tqdm import tqdm

from .search_index import SearchIndex
from ... import utils

logger = logging.getLogger(__name__)


def cosine(s1, s2, **kwargs):
    s1, s2 = set(s1), set(s2)
    return len(s1 & s2) / (math.sqrt(len(s1)) * math.sqrt(len(s2)))


def containment(s1, s2, **kwargs):
    s1, s2 = set(s1), set(s2)
    return len(s1 & s2) / (len(s1) or 1)


def containment_min(s1, s2, **kwargs):
    s1, s2 = set(s1), set(s2)
    return len(s1 & s2) / max(len(s1), len(s2))


def jaccard(s1, s2, **kwargs):
    s1, s2 = set(s1), set(s2)
    return len(s1.intersection(s2)) / len(s1.union(s2))


def weighted_jaccard(s1, s2, **kwargs):
    den = num = 0
    for w in set(s1).union(s2):
        c1, c2 = s1.get(w, 0), s2.get(w, 0)
        num += min(c1, c2)
        den += max(c1, c2)
    return num / den


def weighted_containment(s1, s2, **kwargs):
    den = num = 0
    for w in set(s1).union(s2):
        c1, c2 = s1.get(w, 0), s2.get(w, 0)
        num += min(c1, c2)
        den += c2
    return num / den


class Workload:
    def __init__(self, search_index, queries):
        self.search_index = search_index
        self.queries = queries

    def __call__(self, chunk):
        output = []
        for i in chunk:
            result = self.search_index.query(self.queries[i])
            output.append((i, result))
        return output


def parallel_search(search_index, queries, processes):
    workload = Workload(search_index, queries)
    rows, cols, data = [], [], []

    with mp.Pool(processes) as pool:
        chunks = list(utils.chunks(
            range(len(queries)),
            # make sure to use at most `processes` chunks
            len(queries) // (processes + 1)))
        for idx, chunk in enumerate(pool.map(workload, chunks)):
            for i, results in chunk:
                for j, sim in results:
                    rows.append(i)
                    cols.append(j)
                    data.append(sim)

    return rows, cols, data


class SetSimilarity:
    """
    Approximate set similarity

    similarity_fn : one of ...
    """
    def __init__(self, threshold, similarity_fn='containment'):
        self.threshold = threshold
        self.similarity_fn = similarity_fn
        self.search_index = None

    def fit(self, index, queries=None):
        self.search_index = SearchIndex(
            index, queries=queries,
            similarity_func_name=self.similarity_fn,
            similarity_threshold=self.threshold)
        return self

    def get_similarities(self, queries, index=None, processes=1):
        self_search = index is None
        if index is None:
            index = queries

        self.fit(index, queries=queries)

        processes = mp.cpu_count() if processes < 0 else processes
        logger.info("Using {} CPUs".format(processes))
        if processes == 1:
            sims = dok_matrix((len(queries), len(index)))
            for idx in tqdm(range(len(queries)), total=len(queries),
                            desc="Set similarity: {}".format(self.similarity_fn)):
                for jdx, sim in self.search_index.query(queries[idx]):
                    sims[idx, jdx] = sim
        else:
            results = parallel_search(self.search_index, queries, processes)
            rows, cols, data = results
            sims = csr_matrix((data, (rows, cols)), shape=(len(queries), len(index)))

        # drop self-similarities
        if self_search:
            sims.setdiag(0)
        # transform to csr
        sims = sims.tocsr()
        sims.eliminate_zeros()

        return sims




from setuptools import setup, find_packages
from Cython.Build import cythonize


setup(
    name='text-reuse-retrieve',
    version='0.1.10',
    description='A Python package to perform Text Reuse Detection',
    url='https://github.com/emanjavacas/retrieve',
    author='Enrique Manjavacas',
    license='MIT',
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    include_package_data=True,
    # ext_modules=cythonize('ext/_align.pyx'), # disabling cython extension for now
    install_requires=[
        'Cython>=0.29.19',
        'numpy>=1.17.5',
        'scipy>=1.4.1',
        'gensim>=3.8.3',
        'tqdm>=4.46.0',
        'numba>=0.49.1',
        'pandas>=1.0.4',
        'scikit_learn>=0.23.2',
        'matplotlib>=3.2.1',
        'Flask>=1.1.2',
        'SetSimilaritySearch>=0.1.7'
        'dataclasses>=0.6',
        # 'pyemd>=0.5.1',
        # 'fastText==0.9.2',
    ],
    extras_require={":python_version<'3.7'": ["importlib_resources"]},
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
)




# for SIF, see: https://github.com/PrincetonML/SIF

import inspect

import numpy as np
from sklearn.decomposition import TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer

from .base import VSM


def remove_pc(W, npc=1):
    pc = TruncatedSVD(n_components=npc, n_iter=7).fit(W).components_
    if npc == 1:
        return W - W.dot(pc.transpose()) * pc
    else:
        return W - W.dot(pc.transpose()).dot(pc)


class SIF(VSM):
    def __init__(self, embs, freqs, npc=1, a=1e-3, **kwargs):
        self.npc = npc
        self.a = a
        # attributes
        self.embs = embs
        self.freqs = freqs
        self.N = sum(freqs.values())

    def get_weight(self, w):
        if w not in self.freqs:
            raise KeyError(w)
        return self.a / (self.a + self.freqs[w] / self.N)

    def transform(self, sents):
        # get average
        output = np.array([
            sum(self.get_weight(w) * self.embs[w] for w in sent) / len(sent)
            for sent in sents])
        output = remove_pc(output, npc=self.npc)
        return output


class TfidfEmbedding(VSM):
    def __init__(self, embs, sents, **kwargs):
        # attributes
        self.embs = embs
        params = set(inspect.signature(TfidfVectorizer).parameters)
        self.tfidf = TfidfVectorizer(
            # overwrite default to avoid ignoring input
            token_pattern=r'\S+',
            **{k: v for k, v in kwargs.items() if k in params})

    def fit(self, sents):
        self.tfidf.fit(' '.join(s) for s in sents)
        return self

    def transform(self, sents):
        X = self.tfidf.transform(sents)
        output = np.zeros((len(sents), ))
        idx2w = self.tfidf.get_feature_names()
        for s_idx in range(len(sents)):
            output[s_idx] = sum(
                weight * self.embs[idx2w[idx]]
                for (_, idx), weight in X[s_idx].todok().items()
            ) / len(sents[s_idx])

        return output


class BOWEmbedding(VSM):
    def __init__(self, embs, **kwargs):
        # attributes
        self.embs = embs

    def transform(self, sents):
        return np.array([sum(self.embs[w] for w in s) / len(s) for s in sents])




import logging
import multiprocessing as mp

import numpy as np
import scipy.sparse
import tqdm


logger = logging.getLogger(__name__)


class Workload:
    def __init__(self, coll1, coll2, field='lemma', **kwargs):
        self.coll1 = coll1
        self.coll2 = coll2
        self.field = field
        self.kwargs = kwargs

    def __call__(self, args):
        # unpack
        i, j = args
        score = self.coll1[i].local_alignment(
            self.coll2[j], field=self.field, only_score=True, **self.kwargs)

        return (i, j), score


def align_collections(queries, index=None, S=None, field=None, processes=1, **kwargs):

    if index is None:
        index = queries

    # get target ids
    if S is not None:
        x, y, _ = scipy.sparse.find(S)
    else:
        x, y = np.meshgrid(np.arange(len(queries)), np.arange(len(index)))
        x, y = x.reshape(-1), y.reshape(-1)

    x, y = x.tolist(), y.tolist()

    sims = scipy.sparse.dok_matrix((len(queries), len(index)))  # sparse output

    processes = mp.cpu_count() if processes < 0 else processes
    logger.info("Using {} CPUs".format(processes))
    if processes == 1:
        for i, j in tqdm.tqdm(zip(x, y), total=len(x), desc='Local alignment'):
            score = queries[i].local_alignment(
                index[j], field=field, only_score=True, **kwargs)
            sims[i, j] = score
    else:
        workload = Workload(queries, index, field=field, **kwargs)
        with mp.Pool(processes) as pool:
            for (i, j), score in pool.map(workload, list(zip(x, y))):
                sims[i, j] = score

    return sims.tocsr()


if __name__ == '__main__':
    import timeit

    from retrieve.corpora import load_vulgate
    from retrieve.data import Criterion, TextPreprocessor, FeatureSelector
    from retrieve.embeddings import Embeddings
    from retrieve.methods import create_embedding_scorer, SetSimilarity

    # load
    vulg = load_vulgate(max_verses=1000)
    # preprocess
    TextPreprocessor().process_collections(vulg, min_n=2, max_n=4)
    # drop features and get vocabulary
    FeatureSelector(vulg).filter_collections(
        vulg, (Criterion.DF >= 2) & (Criterion.FREQ >= 5))
    # get documents
    feats = vulg.get_features(cast=set)
    # set-based similarity
    S = SetSimilarity(0.5, similarity_fn="containment").get_similarities(feats)

    # alignment
    TextPreprocessor().process_collections(vulg)
    vocab = FeatureSelector(vulg).filter_collections(
        vulg, (Criterion.DF >= 2) & (Criterion.FREQ >= 5))
    # load embeddings, make sure S is in same order as vocab
    embs = Embeddings.from_resource(
        'latin.lemma.ft.dim100.mc2.embeddings.gz', vocab=vocab)
    # embedding scorer
    scorer = create_embedding_scorer(embs)

    x, y, _ = scipy.sparse.find(S)
    print("Considering {} comparisons".format(len(x)))
    time = timeit.Timer(lambda: align_collections(vulg, vulg, S=S)).timeit(5)
    print(" - Took", time)




import unittest

from retrieve.data import TextPreprocessor, FeatureSelector, Criterion
from retrieve import utils

from retrieve.corpora import load_vulgate


class TestCriterion(unittest.TestCase):
    def setUp(self):
        collection = load_vulgate()
        stops = utils.Stopwords('latin.stop')
        processor = TextPreprocessor(stopwords=stops, field_regexes={'token': '[a-z]+'})
        processor.process_collections(collection)
        self.fsel = FeatureSelector(collection)

    def test_threshold(self):
        for th_min, th_max in zip(range(1, 1000, 100), range(100, 10000, 1000)):
            vocab = self.fsel.get_vocab(th_min <= Criterion.DF < th_max)
            for ft in vocab:
                self.assertTrue(self.fsel.dfs[self.fsel.features[ft]] >= th_min)
                self.assertTrue(self.fsel.dfs[self.fsel.features[ft]] < th_max)




from .search_index import SearchIndex
from .set_similarity import containment, jaccard, containment_min, cosine
from .set_similarity import weighted_containment, weighted_jaccard
from .set_similarity import SetSimilarity




import matplotlib.pyplot as plt


def plot_precision_recall_curve(metrics, at=20, name=None, ax=None, **kwargs):
    if ax is None:
        fig, ax = plt.subplots()

    line_kwargs = {"drawstyle": "steps-post"}
    if name is not None:
        line_kwargs['label'] = name
    line_kwargs = dict(line_kwargs, **kwargs)

    line = ax.plot(metrics[f'r@{at}'], metrics[f'p@{at}'], **line_kwargs)
    ax.set(xlabel="Recall", ylabel="Precision")

    if "label" in line_kwargs:
        ax.legend(loc='upper right')

    return ax, line
